/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <string.h>
#include <libxml/tree.h>
#include "gui.h"

typedef struct _nzb {
    GList *files;
} NZB;

typedef struct _nzb_file {
    char subject[256];
    char poster[256];
    time_t stamp;
    guint64 filesize;
    GList *groups;
    GList *segments;
} NZBFile;

static NZB *nzbfile = NULL;
static char *xml_parser_errmsg = NULL;
static char *collection_name = NULL;
static gboolean quick_import = FALSE;

void    on_btnNZBImport_clicked(GtkWidget *caller, gpointer data);

static void
xml_parser_error_func (void *ctx, const char *msg, ...)
{
    va_list args;
    char *orig = xml_parser_errmsg;
    char *tmp;

    va_start(args, msg);

    tmp = g_strdup_vprintf(msg, args);
    if (orig) {
        xml_parser_errmsg = g_strdup_printf("%s %s", orig, tmp);
        g_free(orig);
        g_free(tmp);
    } else {
        xml_parser_errmsg = tmp;
    }

    va_end(args);
}

static void
parse_groups(xmlDocPtr doc, xmlNodePtr cur, NZBFile *file)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "group")) {
            xmlChar *group = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            file->groups = g_list_append(file->groups, g_strdup((const gchar *) group));
            xmlFree(group);
        }

        cur = cur->next;
    }
}

static void
parse_segments(xmlDocPtr doc, xmlNodePtr cur, NZBFile *file)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "segment")) {
            NNTPGrabPart *segment = g_malloc0(sizeof(NNTPGrabPart));
            xmlChar *value;
            char *message_id;

            value = xmlGetProp(cur, (const xmlChar *) "bytes");
            segment->size = atol((const char *) value);
            file->filesize += segment->size;
            xmlFree(value);

            value = xmlGetProp(cur, (const xmlChar *) "number");
            segment->partnum = atoi((const char *) value);
            xmlFree(value);

            // NZB files contain a message-id without the brackets. Add them manually
            value = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
            message_id = g_strdup_printf("<%s>", (const gchar *) value);
            strncpy((char*) segment->message_id, message_id, (size_t) sizeof(segment->message_id));
            g_free(message_id);
            xmlFree(value);

            file->segments = g_list_append(file->segments, segment);
        }

        cur = cur->next;
    }
}

static void
parse_file(xmlDocPtr doc, xmlNodePtr cur, NZBFile *file)
{
    while(cur != NULL) {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "groups")) {
            parse_groups(doc, cur->xmlChildrenNode, file);
        }

        if (!xmlStrcmp(cur->name, (const xmlChar *) "segments")) {
            parse_segments(doc, cur->xmlChildrenNode, file);
        }

        cur = cur->next;
    }
}

static NZB *
parse_nzb_file(const char *filename, char **errmsg)
{
    xmlDocPtr    doc;
    xmlNodePtr   cur;
    NZB         *nzb = NULL;
    char        *contents = NULL;
    GError      *err = NULL;

    if (!g_file_get_contents(filename, &contents, NULL, &err)) {
        if (errmsg) {
            *errmsg = g_strdup(err->message);
        }
        g_error_free(err);

        return NULL;
    }

    xmlSetGenericErrorFunc(NULL, xml_parser_error_func);

    if (xml_parser_errmsg) {
        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;
    }

    doc = xmlParseMemory(contents, strlen(contents));
    cur = xmlDocGetRootElement(doc);
    if (cur == NULL) {
        if (errmsg) {
            *errmsg = g_strdup(xml_parser_errmsg);
        }

        g_free(xml_parser_errmsg);
        xml_parser_errmsg = NULL;
        g_free(contents);

        xmlFreeDoc(doc);
        return NULL;
    }

    if (xmlStrcmp(cur->name, (const xmlChar *) "nzb")) {
        if (errmsg) {
            *errmsg = g_strdup("parse_nzb_file: document of the wrong type, root node != nzb");
        }

        g_free(contents);
        xmlFreeDoc(doc);

        return FALSE;
    }

    cur = cur->xmlChildrenNode;
    if (!cur) {
        if (errmsg) {
            *errmsg = g_strdup("parse_nzb_file: no children of root-node!");
        }

        g_free(contents);
        xmlFreeDoc(doc);

        return FALSE;
    }

    nzb = g_malloc0(sizeof(NZB));

    do {
        if (!xmlStrcmp(cur->name, (const xmlChar *) "file")) {
            NZBFile *file = g_malloc0(sizeof(NZBFile));
            xmlChar *value;

            value = xmlGetProp(cur, (const xmlChar *) "subject");
            strncpy((char *) file->subject, (const char *) value, (size_t) sizeof(file->subject));
            xmlFree(value);

            value = xmlGetProp(cur, (const xmlChar *) "poster");
            strncpy((char *) file->poster, (const char *) value, (size_t) sizeof(file->poster));
            xmlFree(value);

            value = xmlGetProp(cur, (const xmlChar *) "date");
            file->stamp = atoi((const char *) value);
            xmlFree(value);

            parse_file(doc, cur->xmlChildrenNode, file);

            nzb->files = g_list_append(nzb->files, file);
        }
    } while ((cur=cur->next));

    xmlFreeDoc(doc);
    g_free(contents);

    return nzb;
}

static void
nzb_file_free(NZB *nzbfile)
{
    GList *list;

    list = nzbfile->files;
    while (list) {
        NZBFile *file = (NZBFile *) list->data;
        GList *list2;

        list2 = file->groups;
        while (list2) {
            g_free(list2->data);
            list2 = g_list_next(list2);
        }

        g_list_free(file->groups);

        list2 = file->segments;
        while (list2) {
            g_free(list2->data);
            list2 = g_list_next(list2);
        }

        g_list_free(file->segments);

        g_free(file);

        list = g_list_next(list);
    }

    g_free(nzbfile);
}

static void
strip_extension_from_collection_name(void)
{
    char *tmp;

    tmp = strstr(collection_name, ".nzb");
    if (tmp) {
        tmp[0] = '\0';
    }

    tmp = strstr(collection_name, ".NZB");
    if (tmp) {
        tmp[0] = '\0';
    }
}

// Callbacks from the main window
void
on_btnImportNZB_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowNZB;
    GtkWidget *treeNZBFiles;
    GtkWidget *filechooserNZB;
    GtkWidget *btnNZBImport;
    GtkTreeModel *model;

    windowNZB = get_widget("windowNZB");
    treeNZBFiles = get_widget("treeNZBFiles");
    filechooserNZB = get_widget("filechooserNZB");
    btnNZBImport = get_widget("btnNZBImport");

    gtk_file_chooser_set_filename(GTK_FILE_CHOOSER(filechooserNZB), "");

    model = gtk_tree_view_get_model(GTK_TREE_VIEW(treeNZBFiles));
    gtk_list_store_clear(GTK_LIST_STORE(model));

    gtk_widget_show(windowNZB);
    gtk_widget_set_sensitive(btnNZBImport, FALSE);

    quick_import = FALSE;
}

void
on_btnQuickImportNZB_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowMain;
    GtkWidget *dialog;
    char *errmsg = NULL;

    windowMain = get_widget("windowMain");

    dialog = gtk_file_chooser_dialog_new ("Open File",
                        GTK_WINDOW(windowMain),
                        GTK_FILE_CHOOSER_ACTION_OPEN,
                        GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
                        GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT,
                        NULL);

    if (gtk_dialog_run (GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
        char *filename;

        filename = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));

        if (nzbfile) {
            nzb_file_free(nzbfile);
        }

        nzbfile = parse_nzb_file(filename, &errmsg);
        if (!nzbfile) {
            if (!errmsg) {
                errmsg = g_strdup_printf(_("Error occured while reading NZB File '%s'"), filename);
            }

            dialog_show(windowMain, errmsg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

            g_free(errmsg);
            g_free(filename);
            gtk_widget_destroy (dialog);

            return;
        }

        collection_name = g_path_get_basename(filename);
        strip_extension_from_collection_name();

        g_free (filename);

        on_btnNZBImport_clicked(caller, data);
    }

    gtk_widget_destroy (dialog);

    quick_import = TRUE;
}

static gboolean
nzb_model_foreach_add_func (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
    NZBFile *file;
    gboolean flag;
    GtkWidget *windowNZB;
    char *errmsg = NULL;

    gtk_tree_model_get(model, iter, 5, &flag, 6, &file, -1);

    if (!flag) {
        return FALSE;
    }

    if (!nntpgrab_schedular_add_task_to_queue(core, collection_name, file->subject, file->poster, file->stamp, file->filesize, file->groups, file->segments, &errmsg)) {
        char *msg;

        if (errmsg) {
            msg = g_strdup_printf(_("File could not be added to the download queue:\n%s"), errmsg);
            g_free(errmsg);
        } else {
            msg = g_strdup_printf(_("File could not be added to the download queue"));
        }

        windowNZB = get_widget("windowNZB");
        dialog_show(windowNZB, msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
        g_free(msg);
    }

    return FALSE;
}

void
on_radioAddToNewCollection_toggled(GtkWidget *caller, gpointer data)
{
    GtkWidget *frameAddToNewCollection;
    GtkWidget *entryCollectionName;
    GtkWidget *frameAddToExistingCollection;
    GtkWidget *comboCollectionName;

    frameAddToNewCollection = get_widget("frameAddToNewCollection");
    entryCollectionName = get_widget("entryCollectionName");
    frameAddToExistingCollection = get_widget("frameAddToExistingCollection");
    comboCollectionName = get_widget("comboCollectionName");

    gtk_widget_set_sensitive(frameAddToNewCollection, TRUE);
    gtk_widget_set_sensitive(frameAddToExistingCollection, FALSE);
}

void
on_radioAddToExistingCollection_toggled(GtkWidget *caller, gpointer data)
{
    GtkWidget *frameAddToNewCollection;
    GtkWidget *entryCollectionName;
    GtkWidget *frameAddToExistingCollection;
    GtkWidget *comboCollectionName;

    frameAddToNewCollection = get_widget("frameAddToNewCollection");
    entryCollectionName = get_widget("entryCollectionName");
    frameAddToExistingCollection = get_widget("frameAddToExistingCollection");
    comboCollectionName = get_widget("comboCollectionName");

    gtk_widget_set_sensitive(frameAddToNewCollection, FALSE);
    gtk_widget_set_sensitive(frameAddToExistingCollection, TRUE);
}

void
on_btnSetCollectionNameCancel_clicked(GtkWidget *caller, gpointer data)
{
    gtk_widget_hide(get_widget("windowSetCollectionName"));
}

void
on_btnSetCollectionNameOK_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowNZB;
    GtkWidget *treeNZBFiles;
    GtkWidget *radioAddToNewCollection;
    GtkWidget *entryCollectionName;
    GtkWidget *comboCollectionName;
    GtkTreeModel *store;
    char *errmsg;

    radioAddToNewCollection = get_widget("radioAddToNewCollection");
    entryCollectionName = get_widget("entryCollectionName");
    comboCollectionName = get_widget("comboCollectionName");

    // Set the right collection name
    g_free(collection_name);
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(radioAddToNewCollection))) {
        collection_name = g_strdup(gtk_entry_get_text(GTK_ENTRY(entryCollectionName)));
    } else {
        GtkTreeIter iter;

        if (!gtk_combo_box_get_active_iter(GTK_COMBO_BOX(comboCollectionName), &iter)) {
            // No item selected, use the entry field
            collection_name = g_strdup(gtk_entry_get_text(GTK_ENTRY(entryCollectionName)));
        } else {
            store = gtk_combo_box_get_model(GTK_COMBO_BOX(comboCollectionName));
            gtk_tree_model_get(store, &iter, 0, &collection_name, -1);
        }
    }

    if (quick_import) {
        GList *list = nzbfile->files;
        while (list) {
            NZBFile *file = list->data;

            if (!nntpgrab_schedular_add_task_to_queue(core, collection_name, file->subject, file->poster, file->stamp, file->filesize, file->groups, file->segments, &errmsg)) {
                GtkWidget *windowMain;
                char *msg;

                windowMain = get_widget("windowMain");

                if (errmsg) {
                    msg = g_strdup_printf(_("File could not be added to the download queue:\n%s"), errmsg);
                    g_free(errmsg);
                } else {
                    msg = g_strdup_printf(_("File could not be added to the download queue"));
                }

                dialog_show(windowMain, msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
                g_free(msg);
            }

            list = g_list_next(list);
        }

        g_free(collection_name);
        collection_name = NULL;

        nzb_file_free(nzbfile);
        nzbfile = NULL;
    } else {
        treeNZBFiles = get_widget("treeNZBFiles");
        store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeNZBFiles));

        gtk_tree_model_foreach(store, nzb_model_foreach_add_func, GINT_TO_POINTER(TRUE));
    }

    windowNZB = get_widget("windowNZB");

    if (!nntpgrab_schedular_save_queue(core, &errmsg)) {
        char *msg;
        if (errmsg) {
            msg = g_strdup_printf(_("Download queue could not be saved:\n%s"), errmsg);
            g_free(errmsg);
        } else {
            msg = g_strdup_printf(_("Download queue could not be saved"));
        }

        dialog_show(windowNZB, msg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);
        g_free(msg);
    }

    gtk_widget_hide(windowNZB);

    if (nzbfile) {
        nzb_file_free(nzbfile);
        nzbfile = NULL;
    }

    if (collection_name) {
        g_free(collection_name);
        collection_name = NULL;
    }

    gtk_widget_hide(get_widget("windowSetCollectionName"));
}

// Callbacks from the NZB Import window
void
on_filechooserNZB_file_set(GtkWidget *caller, gpointer data)
{
    char *filename;
    char *errmsg = NULL;
    GtkWidget *windowNZB;
    GtkWidget *btnNZBImport;
    GtkWidget *treeNZBFiles;
    GtkTreeModel *store;
    GList *list;

    windowNZB = get_widget("windowNZB");
    btnNZBImport = get_widget("btnNZBImport");
    filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(caller));

    if (!filename) {
        return;
    }

    if (nzbfile) {
        nzb_file_free(nzbfile);
    }

    nzbfile = parse_nzb_file(filename, &errmsg);
    if (!nzbfile) {
        if (!errmsg) {
            errmsg = g_strdup_printf(_("Error occured while reading NZB File '%s'"), filename);
        }

        dialog_show(windowNZB, errmsg, GTK_MESSAGE_ERROR, GTK_BUTTONS_OK);

        g_free(errmsg);
        return;
    }

    collection_name = g_path_get_basename(filename);
    strip_extension_from_collection_name();

    g_free(filename);

    treeNZBFiles = get_widget("treeNZBFiles");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeNZBFiles));

    gtk_list_store_clear(GTK_LIST_STORE(store));

    list = nzbfile->files;
    while (list) {
        NZBFile *file = list->data;
        GtkTreeIter iter;

        gtk_list_store_append(GTK_LIST_STORE(store), &iter);

        gtk_list_store_set(GTK_LIST_STORE(store), &iter, 0, file->subject, 1, file->poster, 2, file->groups->data, 3, file->stamp, 4, file->filesize, 5, FALSE, 6, file, -1);

        list = g_list_next(list);
    }

    gtk_widget_set_sensitive(btnNZBImport, TRUE);
}

static gboolean
nzb_model_foreach_select_func (GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data)
{
    gtk_list_store_set(GTK_LIST_STORE(model), iter, 5, GPOINTER_TO_INT(data), -1);

    return FALSE;
}

void
on_btnNZBSelectNone_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *treeNZBFiles;
    GtkTreeModel *store;

    treeNZBFiles = get_widget("treeNZBFiles");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeNZBFiles));

    gtk_tree_model_foreach(store, nzb_model_foreach_select_func, GINT_TO_POINTER(FALSE));
}

void
on_btnNZBSelectAll_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *treeNZBFiles;
    GtkTreeModel *store;

    treeNZBFiles = get_widget("treeNZBFiles");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeNZBFiles));

    gtk_tree_model_foreach(store, nzb_model_foreach_select_func, GINT_TO_POINTER(TRUE));
}

void
on_btnNZBCancel_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowNZB;

    windowNZB = get_widget("windowNZB");
    gtk_widget_hide(windowNZB);

    if (nzbfile) {
        nzb_file_free(nzbfile);
        nzbfile = NULL;
    }

    if (collection_name) {
        g_free(collection_name);
        collection_name = NULL;
    }
}

static void
foreach_collection_func(const char *collection_name, guint64 total_size, gpointer data)
{
    GtkTreeIter iter;

    gtk_list_store_append(GTK_LIST_STORE(data), &iter);
    gtk_list_store_set(GTK_LIST_STORE(data), &iter, 0, collection_name, -1);
}

void
on_btnNZBImport_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *windowSetCollectionName;
    GtkWidget *windowMain;
    GtkWidget *windowNZB;
    GtkWidget *frameAddToNewCollection;
    GtkWidget *entryCollectionName;
    GtkWidget *frameAddToExistingCollection;
    GtkWidget *comboCollectionName;
    GtkWidget *radioAddToNewCollection;
    GtkWidget *radioAddToExistingCollection;
    GtkTreeModel *store;

    windowSetCollectionName = get_widget("windowSetCollectionName");
    windowMain = get_widget("windowMain");
    windowNZB = get_widget("windowNZB");
    frameAddToNewCollection = get_widget("frameAddToNewCollection");
    entryCollectionName = get_widget("entryCollectionName");
    frameAddToExistingCollection = get_widget("frameAddToExistingCollection");
    comboCollectionName = get_widget("comboCollectionName");
    radioAddToNewCollection = get_widget("radioAddToNewCollection");
    radioAddToExistingCollection = get_widget("radioAddToExistingCollection");

    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radioAddToNewCollection), TRUE);
    gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(radioAddToExistingCollection), FALSE);

    gtk_widget_set_sensitive(frameAddToNewCollection, TRUE);
    gtk_widget_set_sensitive(frameAddToExistingCollection, FALSE);
    gtk_entry_set_text(GTK_ENTRY(entryCollectionName), collection_name);

    // Clear the contents of the combo box
    store = gtk_combo_box_get_model(GTK_COMBO_BOX(comboCollectionName));
    gtk_list_store_clear(GTK_LIST_STORE(store));

    nntpgrab_schedular_foreach_task(core, foreach_collection_func, NULL, NULL, store);
    gtk_combo_box_set_active (GTK_COMBO_BOX (comboCollectionName), 0);

    // FIXME: Due to a possible bug in GTK it isn't possible to set another transient window
    // when one is already attached
//    if (quick_import) {
        gtk_window_set_transient_for(GTK_WINDOW(windowSetCollectionName), GTK_WINDOW(windowMain));
//    } else {
//        gtk_window_set_transient_for(GTK_WINDOW(windowSetCollectionName), GTK_WINDOW(windowNZB));
//    }

    gtk_widget_show(windowSetCollectionName);
    gtk_window_present(GTK_WINDOW(windowSetCollectionName));
}

gboolean
on_windowNZB_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    gtk_widget_hide(widget);

    if (nzbfile) {
        nzb_file_free(nzbfile);
        nzbfile = NULL;
    }

    if (collection_name) {
        g_free(collection_name);
        collection_name = NULL;
    }

    return TRUE;
}

gboolean
on_windowSetCollectionName_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    gtk_widget_hide(widget);

    return TRUE;
}

static void
download_check_toggled(GtkWidget *caller, gchar *path, gpointer data)
{
    gboolean value;
    GtkTreeIter iter;
    GtkWidget *treeNZBFiles;
    GtkTreeModel *model;

    treeNZBFiles = get_widget("treeNZBFiles");
    model = gtk_tree_view_get_model(GTK_TREE_VIEW(treeNZBFiles));

    if (gtk_tree_model_get_iter_from_string(model, &iter, path)) {
        gtk_tree_model_get(model, &iter, 5, &value, -1);
        if (value) {
            value = FALSE;
        } else {
            value = TRUE;
        }
        gtk_list_store_set(GTK_LIST_STORE(model), &iter, 5, value, -1);
    }
}

void
nzb_initialize()
{
    GtkWidget *windowNZB;
    GtkWidget *treeNZBFiles;
    GtkTreeSelection *selection;
    GtkListStore *store;
    GtkCellRenderer *cellRenderer;
    GtkWidget *comboCollectionName;

    windowNZB = get_widget("windowNZB");
    treeNZBFiles = get_widget("treeNZBFiles");
    comboCollectionName = get_widget("comboCollectionName");

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeNZBFiles));
    gtk_tree_selection_set_mode(selection, GTK_SELECTION_MULTIPLE);

    cellRenderer = gtk_cell_renderer_toggle_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeNZBFiles), -1, _("Download"), cellRenderer, "active", 5, NULL);
    g_signal_connect((gpointer) cellRenderer, "toggled", G_CALLBACK(download_check_toggled), NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeNZBFiles), -1, _("Subject"), cellRenderer, "text", 0, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeNZBFiles), -1, _("Poster"), cellRenderer, "text", 1, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeNZBFiles), -1, _("Groups"), cellRenderer, "text", 2, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeNZBFiles), -1, _("Stamp"), cellRenderer, "text", 3, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeNZBFiles), -1, _("File size"), cellRenderer, "text", 4, NULL);

    store = gtk_list_store_new(7, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_UINT64, G_TYPE_BOOLEAN, G_TYPE_POINTER);

    gtk_tree_view_set_model(GTK_TREE_VIEW(treeNZBFiles), GTK_TREE_MODEL(store));

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (comboCollectionName), cellRenderer, FALSE);
    gtk_cell_layout_set_attributes (GTK_CELL_LAYOUT (comboCollectionName), cellRenderer, "text", 0, NULL);

    store = gtk_list_store_new(1, G_TYPE_STRING);
    gtk_combo_box_set_model(GTK_COMBO_BOX(comboCollectionName), GTK_TREE_MODEL(store));
}
