/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef WIN32
#include <windows.h>
#include <gdk/gdkwin32.h>
#endif
#include <string.h>
#include "gui.h"

static gboolean find_iter_by_collection_name(GtkTreeModel *model, const char *collection_name, GtkTreeIter *iter);
static gboolean find_iter_by_subject(GtkTreeModel *model, GtkTreeIter parent, const char *subject, GtkTreeIter *iter);

static SchedularState status_flag = SCHEDULAR_STATE_RUNNING;     // Schedular is running by default

enum {
    FIELD_COLLECTION_NAME,
    FIELD_SUBJECT,
    FIELD_POSTER,
    FIELD_GROUPS,
    FIELD_FILESIZE_STR,
    FIELD_STAMP_STR,
    FIELD_PROGRESS,
    FIELD_FILESIZE,
    FIELD_NUM_PARTS,
    FIELD_PARTS_DONE,
    FIELD_PARTS_FAILED,
    FIELD_DECODE_PROGRESS_STR,
    LAST_FIELD
};

#ifdef WIN32
static void
open_file(const char *file, gboolean open_folder)
{
    char *mode;
    GtkWidget *windowMain = get_widget("windowMain");

    if (open_folder) {
        mode = "explore";
    } else {
        mode = "open";
    }

    if ((int) ShellExecuteA((HWND) GDK_WINDOW_HWND(windowMain->window), mode, file, NULL, file, SW_SHOW) < 32) {
        char *msg;
        GtkWidget *windowMain;

        windowMain = get_widget("windowMain");
        if (open_folder) {
            msg = g_strdup_printf(_("Unable to navigate to the download folder\n"));    // TODO: Add GetLastError message
        } else {
            msg = g_strdup_printf(_("Unable to execute file\n%s"), file);    // TODO: Add GetLastError message
        }
        dialog_show(windowMain, msg, GTK_MESSAGE_INFO, GTK_BUTTONS_OK);
        g_free(msg);
    }
}
#else
static void
open_file(const char *folder_unquoted, gboolean open_folder)
{
    char *cmd;
    GError *err = NULL;
    char *folder = g_shell_quote(folder_unquoted);

#ifdef DARWIN
    cmd = g_strdup_printf("open %s", folder);
#else
    cmd = g_strdup_printf("xdg-open %s", folder);
#endif

    if (!g_spawn_command_line_async(cmd, &err)) {
        char *msg;
        GtkWidget *windowMain;

        windowMain = get_widget("windowMain");
        if (open_folder) {
            msg = g_strdup_printf(_("Unable to navigate to the download folder:\n%s"), err->message);
        } else {
            msg = g_strdup_printf(_("Unable to execute file\n%s:\n%s"), folder, err->message);
        }
        g_error_free(err);
        dialog_show(windowMain, msg, GTK_MESSAGE_INFO, GTK_BUTTONS_OK);
        g_free(msg);
    }

    g_free(cmd);
}
#endif

gboolean
on_treeDownloadQueue_button_press_event(GtkWidget *caller, GdkEventButton *event, gpointer data)
{
    GtkWidget *menuDownloadQueue;
    GtkTreeModel *model;
    GtkTreeSelection *selection;
#if GTK_CHECK_VERSION(2,12,0)
    GObject *menuitemExecuteItem;
    GValue val;
#else
    GtkWidget *menuitemExecuteItem;
#endif

    // We only want to catch the right mouse button
    if (event->button != 3) {
        return FALSE;
    }

    // Is at least one item selected?
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(caller));
    if (gtk_tree_selection_count_selected_rows(selection) == 0)
    {
        GtkTreePath *path;

        /* Get tree path for row that was clicked */
        if (!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(caller),
                                            (gint) event->x,
                                            (gint) event->y,
                                            &path, NULL, NULL, NULL))
        {
            // No item(s) selected and no item under the cursor. Ignore
            return FALSE;
        }

        gtk_tree_selection_unselect_all(selection);
        gtk_tree_selection_select_path(selection, path);
        gtk_tree_path_free(path);
    }

    // Do we have exactly one selected item and is this item already downloaded?
#if GTK_CHECK_VERSION(2,12,0)
    menuitemExecuteItem = get_object("menuitemExecuteItem");
    memset(&val, 0, sizeof(val));
    g_value_init(&val, G_TYPE_BOOLEAN);
    g_value_set_boolean (&val, FALSE);
    g_object_set_property(G_OBJECT(menuitemExecuteItem), "sensitive", &val);
#else
    menuitemExecuteItem = get_widget("menuitemExecuteItem");
    gtk_widget_set_sensitive(menuitemExecuteItem, FALSE);
#endif
    if (gtk_tree_selection_count_selected_rows(selection) == 1) {
        char *progress_str;
        GtkTreeIter iter;
        GList *list = gtk_tree_selection_get_selected_rows(selection, &model);

        gtk_tree_model_get_iter(model, &iter, (GtkTreePath *) list->data);
        gtk_tree_model_get(model, &iter, FIELD_DECODE_PROGRESS_STR, &progress_str, -1);

        if (progress_str != NULL && !strcmp(progress_str, _("Done"))) {
#if GTK_CHECK_VERSION(2,12,0)
            g_value_set_boolean (&val, TRUE);
            g_object_set_property(G_OBJECT(menuitemExecuteItem), "sensitive", &val);
#else
            gtk_widget_set_sensitive(menuitemExecuteItem, TRUE);
#endif
        }

        g_list_foreach (list, (GFunc) gtk_tree_path_free, NULL);
        g_list_free (list);
    }

    menuDownloadQueue = get_widget("menuDownloadQueue");
    gtk_menu_popup(GTK_MENU(menuDownloadQueue), NULL, NULL, NULL, NULL, event->button, event->time);

    return TRUE;
}

void
on_menuitemRestartTask_activate(GtkWidget *caller, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows;
    GtkTreeModel *model;
    GtkTreeIter iter, parent;
    char *collection_name, *subject;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_assert(gtk_tree_selection_count_selected_rows(selection) > 0);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    // Duplicate the list and transform the items to GtkTreeRowReferences
    // This is because we are going to delete items in the tree
    list = selectedRows;
    while (list) {
        GtkTreePath *path = list->data;

        gtk_tree_model_get_iter(model, &iter, path);
        if (!gtk_tree_model_iter_parent(model, &parent, &iter)) {
            // Item contains no parent, restart the whole collection
            gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &collection_name, -1);
            nntpgrab_schedular_restart_task(core, collection_name, NULL, NULL);

            g_free(collection_name);
        } else {
            // Restart one item from the collection
            gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);
            gtk_tree_model_get(model, &iter, FIELD_SUBJECT, &subject, -1);

            nntpgrab_schedular_restart_task(core, collection_name, subject, NULL);

            g_free(collection_name);
            g_free(subject);
        }

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);
}

void
on_menuitemOpenDownloadFolder_activate(GtkWidget *caller, gpointer data)
{
    // Is only one collection selected or more ?
    gboolean multiple_collections = FALSE;
    char *collection_name = NULL;
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows;
    GtkTreeModel *model;
    ConfigOpts opts;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_assert(gtk_tree_selection_count_selected_rows(selection) > 0);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    list = selectedRows;
    while (list) {
        GtkTreePath *path = list->data;
        GtkTreeIter iter;
        char *name;

        gtk_tree_model_get_iter(model, &iter, path);
        gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &name, -1);

        if (name == NULL || strlen(name) == 0) {
            GtkTreeIter parent;

            if (name) {
                g_free(name);
            }

            // Look at the parent
            gtk_tree_model_iter_parent(model, &parent, &iter);
            gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &name, -1);
        }

        g_assert(name);

        if (!collection_name) {
            collection_name = g_strdup(name);
        } else {
            if (strcmp(name, collection_name)) {
                // A previous found collection name and this collection name differ!
                multiple_collections = TRUE;
            }
        }

        g_free(name);

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);

    opts = nntpgrab_config_get_opts(core);

    if (multiple_collections) {
        // Open the default download folder
        open_file(opts.download_directory, TRUE);
    } else {
        // Open the download folder of the active collection
        char *folder = g_strdup_printf("%s%s%s", opts.download_directory, G_DIR_SEPARATOR_S, collection_name);

        if (!g_file_test(folder, G_FILE_TEST_IS_DIR)) {
            // Folder doesn't exist yet, fall back to default download directory
            open_file(opts.download_directory, TRUE);
        } else {
            open_file(folder, TRUE);
        }

        g_free(folder);
    }

    if (collection_name) {
        g_free(collection_name);
    }
}

void
on_menuitemExecuteItem_activate(GtkWidget *caller, gpointer data)
{
    char *collection_name = NULL;
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *selectedRows;
    GtkTreeModel *model;
    ConfigOpts opts;
    GtkTreePath *path;
    GtkTreeIter iter;
    char *subject;
    char *filename;
    char *full_filename;
    GtkTreeIter parent;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_assert(gtk_tree_selection_count_selected_rows(selection) == 1);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    path = selectedRows->data;

    gtk_tree_model_get_iter(model, &iter, path);
    gtk_tree_model_get(model, &iter, FIELD_SUBJECT, &subject, -1);

    // Look at the parent
    gtk_tree_model_iter_parent(model, &parent, &iter);
    gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);

    opts = nntpgrab_config_get_opts(core);
    if (!nntpgrab_utils_strip_subject(subject, NULL, NULL, NULL, &filename, NULL, NULL, NULL, NULL)) {
        char *msg = g_strdup(_("Unable to find out the name of the file"));
        GtkWidget *windowMain = get_widget("windowMain");

        dialog_show(windowMain, msg, GTK_MESSAGE_INFO, GTK_BUTTONS_OK);

        g_free(msg);
    } else {
        full_filename = g_strdup_printf("%s%s%s%s%s", opts.download_directory, G_DIR_SEPARATOR_S, collection_name, G_DIR_SEPARATOR_S, filename);

        open_file(full_filename, FALSE);

        g_free(full_filename);
        g_free(filename);
    }

    g_free(subject);
    g_free(collection_name);
    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);
}

void
on_btnRemoveItemFromQueue_clicked(GtkWidget *caller, gpointer data)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeSelection *selection;
    GList *list, *selectedRows, *list_new;
    GtkTreeModel *model;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));

    g_assert(gtk_tree_selection_count_selected_rows(selection) > 0);

    selectedRows = gtk_tree_selection_get_selected_rows(GTK_TREE_SELECTION(selection), &model);

    // Duplicate the list and transform the items to GtkTreeRowReferences
    // This is because we are going to delete items in the tree
    list = selectedRows;
    list_new = NULL;
    while (list) {
        GtkTreePath *path = list->data;

        list_new = g_list_append(list_new, gtk_tree_row_reference_new(model, path));

        list = g_list_next(list);
    }

    g_list_foreach (selectedRows, (GFunc) gtk_tree_path_free, NULL);
    g_list_free (selectedRows);

    // Now loop trough all the items and remove the downloads tasks
    list = list_new;
    while (list) {
        GtkTreeRowReference *ref = list->data;

        if (gtk_tree_row_reference_valid(ref)) {
            GtkTreePath *path = gtk_tree_row_reference_get_path(ref);
            GtkTreeIter iter, parent;
            char *collection_name;
            char *subject;

            gtk_tree_model_get_iter(model, &iter, path);

            if (!gtk_tree_model_iter_parent(model, &parent, &iter)) {
                // Item contains no parent, delete the whole collection
                gtk_tree_model_get(model, &iter, FIELD_COLLECTION_NAME, &collection_name, -1);
                nntpgrab_schedular_del_task_from_queue(core, collection_name, NULL, NULL);

                g_free(collection_name);
            } else {
                // Delete one item from the collection
                gtk_tree_model_get(model, &parent, FIELD_COLLECTION_NAME, &collection_name, -1);
                gtk_tree_model_get(model, &iter, FIELD_SUBJECT, &subject, -1);

                nntpgrab_schedular_del_task_from_queue(core, collection_name, subject, NULL);

                g_free(collection_name);
                g_free(subject);
            }
        }

        gtk_tree_row_reference_free(ref);

        list = g_list_next(list);
    }

    g_list_free(list_new);
}

static void
file_download_done (NNTPGrabCore *core, const char *collection_name, const char *subject)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_PROGRESS, 100, FIELD_DECODE_PROGRESS_STR, _("Waiting"), -1);
}

static void
file_decode_begin (NNTPGrabCore *core, const char *collection_name, const char *subject)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_DECODE_PROGRESS_STR, _("Decoding..."), -1);
}

static void
file_decode_done (NNTPGrabCore *core, const char *collection_name, const char *subject)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_DECODE_PROGRESS_STR, _("Done"), -1);
}

static char *
generate_stamp(time_t stamp)
{
    char buf[64];
    struct tm *now;

    if ((now = localtime(&stamp)) == NULL) {
        // Date could not be parted
        return g_strdup("");
    }

    strftime(buf, sizeof(buf), "%c", now);

    return g_strdup(buf);
}

static int
calculate_progress(int parts_total, int parts_done, int parts_failed)
{
    double progress;
    int progress_int;

    // Prevent a division by zero
    if (parts_total == 0) {
        parts_total = 1;
    }

    progress = parts_done + parts_failed;
    progress /= parts_total;
    progress *= 100;
    progress_int = (int) progress;

    if (progress_int > 100) {
        g_print(__FILE__ ":%i progress overflow!\n", __LINE__);
        progress_int = 100;
    }

    return progress_int;
}

static void
part_done (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int progress;
    int parts_total, parts_done, parts_failed;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    // Update the parent
    gtk_tree_model_get(GTK_TREE_MODEL(store), &parent, FIELD_NUM_PARTS, &parts_total, FIELD_PARTS_DONE, &parts_done, FIELD_PARTS_FAILED, &parts_failed, -1);
    parts_done++;

    progress = calculate_progress(parts_total, parts_done, parts_failed);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent, FIELD_PROGRESS, progress, FIELD_PARTS_DONE, parts_done, -1);

    // Update the child
    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, FIELD_NUM_PARTS, &parts_total, FIELD_PARTS_DONE, &parts_done, FIELD_PARTS_FAILED, &parts_failed, -1);
    parts_done++;

    progress = calculate_progress(parts_total, parts_done, parts_failed);

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_PROGRESS, progress, FIELD_PARTS_DONE, parts_done, -1);
}

static void
part_failed (NNTPGrabCore *core, const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int parts_total, parts_done, parts_failed;
    int progress;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    // Update the parent
    gtk_tree_model_get(GTK_TREE_MODEL(store), &parent, FIELD_NUM_PARTS, &parts_total, FIELD_PARTS_DONE, &parts_done, FIELD_PARTS_FAILED, &parts_failed, -1);
    parts_failed++;

    progress = calculate_progress(parts_total, parts_done, parts_failed);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent, FIELD_PROGRESS, progress, FIELD_PARTS_FAILED, parts_failed, -1);

    // Update the child
    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, FIELD_NUM_PARTS, &parts_total, FIELD_PARTS_DONE, &parts_done, FIELD_PARTS_FAILED, &parts_failed, -1);
    parts_failed++;

    progress = calculate_progress(parts_total, parts_done, parts_failed);

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_PROGRESS, progress, FIELD_PARTS_FAILED, parts_failed, -1);
}

static void
collection_added (NNTPGrabCore *core, const char *collection_name)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    gtk_tree_store_append(GTK_TREE_STORE(store), &iter, NULL);
    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_COLLECTION_NAME, collection_name, FIELD_SUBJECT, collection_name, FIELD_DECODE_PROGRESS_STR, NULL, -1);
}

static void
collection_removed (NNTPGrabCore *core, const char *collection_name)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(store, collection_name, &iter));

    gtk_tree_store_remove(GTK_TREE_STORE(store), &iter);
}

static void
file_added (NNTPGrabCore *core, const char *collection_name, const char *subject, const char *poster, time_t stamp, guint64 file_size, int num_parts, GList *groups)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter parent, iter;
    char *size;
    char *total_groups = NULL;
    char *stamp_str;
    GList *list;
    guint64 file_size_total;
    int num_parts_total, num_parts_done_total, num_parts_failed_total;
    int progress;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));

    list = groups;
    while (list) {
        if (total_groups) {
            char *tmp = g_strdup_printf("%s, %s", total_groups, (const char *) list->data);
            g_free(total_groups);
            total_groups = tmp;
        } else {
            total_groups = g_strdup(list->data);
        }

        list = g_list_next(list);
    }

    size = nntpgrab_utils_calculate_file_size(file_size);
    stamp_str = generate_stamp(stamp);
    gtk_tree_store_append(GTK_TREE_STORE(store), &iter, &parent);
    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_SUBJECT, subject, FIELD_POSTER, poster, FIELD_GROUPS, total_groups, FIELD_FILESIZE_STR, size, FIELD_STAMP_STR, stamp_str, FIELD_FILESIZE, file_size, FIELD_NUM_PARTS, num_parts, FIELD_DECODE_PROGRESS_STR, NULL, -1);

    g_free(stamp_str);
    g_free(size);
    g_free(total_groups);

    gtk_tree_model_get(GTK_TREE_MODEL(store), &parent, FIELD_FILESIZE, &file_size_total, FIELD_NUM_PARTS, &num_parts_total, FIELD_PARTS_DONE, &num_parts_done_total, FIELD_PARTS_FAILED, &num_parts_failed_total, -1);

    file_size_total += file_size;
    num_parts_total += num_parts;
    size = nntpgrab_utils_calculate_file_size(file_size_total);

    progress = calculate_progress(num_parts_total, num_parts_done_total, num_parts_failed_total);
    gtk_tree_store_set(GTK_TREE_STORE(store), &parent, FIELD_FILESIZE_STR, size, FIELD_FILESIZE, file_size_total, FIELD_NUM_PARTS, num_parts_total, FIELD_PROGRESS, progress, -1);

    g_free(size);
}

static void
file_removed (NNTPGrabCore *core, const char *collection_name, const char *subject)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int progress;
    guint64 file_size, file_size_total;
    int num_parts, num_parts_total;
    int num_parts_done, num_parts_done_total;
    int num_parts_failed, num_parts_failed_total;
    char *size;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, FIELD_FILESIZE, &file_size, FIELD_NUM_PARTS, &num_parts, FIELD_PARTS_DONE, &num_parts_done, FIELD_PARTS_FAILED, &num_parts_failed, -1);
    gtk_tree_model_get(GTK_TREE_MODEL(store), &parent, FIELD_FILESIZE, &file_size_total, FIELD_NUM_PARTS, &num_parts_total, FIELD_PARTS_DONE, &num_parts_done_total, FIELD_PARTS_FAILED, &num_parts_failed_total, -1);

    file_size_total -= file_size;
    num_parts_total -= num_parts;
    num_parts_done_total -= num_parts_done;
    num_parts_failed_total -= num_parts_failed;
    size = nntpgrab_utils_calculate_file_size(file_size_total);

    progress = calculate_progress(num_parts_total, num_parts_done_total, num_parts_failed_total);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent, FIELD_FILESIZE_STR, size, FIELD_FILESIZE, file_size_total, FIELD_NUM_PARTS, num_parts_total, FIELD_PARTS_DONE, num_parts_done_total, FIELD_PARTS_FAILED, num_parts_failed_total, FIELD_PROGRESS, progress, -1);

    gtk_tree_store_remove(GTK_TREE_STORE(store), &iter);

    g_free(size);
}

static void
file_restarted (NNTPGrabCore *core, const char *collection_name, const char *subject)
{
    GtkWidget *treeDownloadQueue;
    GtkTreeModel *store;
    GtkTreeIter iter, parent;
    int progress;
    int num_parts, num_parts_total;
    int num_parts_done, num_parts_done_total;
    int num_parts_failed, num_parts_failed_total;

    treeDownloadQueue = get_widget("treeDownloadQueue");
    store = gtk_tree_view_get_model(GTK_TREE_VIEW(treeDownloadQueue));

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    // Update the row itself
    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, FIELD_NUM_PARTS, &num_parts, FIELD_PARTS_DONE, &num_parts_done, FIELD_PARTS_FAILED, &num_parts_failed, -1);

    gtk_tree_store_set(GTK_TREE_STORE(store), &iter, FIELD_PARTS_DONE, 0, FIELD_PARTS_FAILED, 0, FIELD_PROGRESS, 0, FIELD_DECODE_PROGRESS_STR, NULL, -1);

    // Update the parent
    gtk_tree_model_get(GTK_TREE_MODEL(store), &parent, FIELD_NUM_PARTS, &num_parts_total, FIELD_PARTS_DONE, &num_parts_done_total, FIELD_PARTS_FAILED, &num_parts_failed_total, -1);

    num_parts_total -= num_parts;
    num_parts_done_total -= num_parts_done;
    num_parts_failed_total -= num_parts_failed;

    progress = calculate_progress(num_parts_total, num_parts_done_total, num_parts_failed_total);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent, FIELD_PARTS_DONE, num_parts_done_total, FIELD_PARTS_FAILED, num_parts_failed_total, FIELD_PROGRESS, progress, -1);
}

static void
foreach_collection_func (const char *collection_name, guint64 total_size, gpointer data)
{
    GtkTreeStore *store = GTK_TREE_STORE(data);
    GtkTreeIter iter;
    char *size;

    size = nntpgrab_utils_calculate_file_size(total_size);
    gtk_tree_store_append(store, &iter, NULL);
    gtk_tree_store_set(store, &iter, FIELD_COLLECTION_NAME, collection_name, FIELD_SUBJECT, collection_name, FIELD_FILESIZE_STR, size, FIELD_FILESIZE, total_size, FIELD_DECODE_PROGRESS_STR, NULL, -1);
    g_free(size);
}

static gboolean
find_iter_by_collection_name(GtkTreeModel *model, const char *collection_name, GtkTreeIter *iter)
{
    char *name;

    if (gtk_tree_model_get_iter_first(GTK_TREE_MODEL(model), iter)) {
        do {
            gtk_tree_model_get(GTK_TREE_MODEL(model), iter, FIELD_COLLECTION_NAME, &name, -1);

            if (!strcmp(collection_name, name)) {
                g_free(name);
                return TRUE;
            }

            g_free(name);
        } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(model), iter));
    }

    return FALSE;
}

static gboolean
find_iter_by_subject(GtkTreeModel *model, GtkTreeIter parent, const char *subject, GtkTreeIter *iter)
{
    char *name;

    if (!gtk_tree_model_iter_children(GTK_TREE_MODEL(model), iter, &parent)) {
        // Node contains no children. Shouldn't happen
        return FALSE;
    }

    do {
        gtk_tree_model_get(GTK_TREE_MODEL(model), iter, FIELD_SUBJECT, &name, -1);

        if (!strcmp(subject, name)) {
            g_free(name);
            return TRUE;
        }

        g_free(name);
    } while (gtk_tree_model_iter_next(GTK_TREE_MODEL(model), iter));

    return FALSE;
}

static void
foreach_file_func (const char *collection_name, const char *subject, const char *poster, time_t stamp, guint64 file_size, int position, int num_parts, int num_parts_downloaded, int num_parts_failed, gpointer data)
{
    GtkTreeStore *store = GTK_TREE_STORE(data);
    GtkTreeIter iter, parent;
    char *size;
    char *stamp_str;
    int num_parts_total;
    int num_parts_downloaded_total;
    int num_parts_failed_total;
    int progress;
    char *progress_str;

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));

    size = nntpgrab_utils_calculate_file_size(file_size);
    stamp_str = generate_stamp(stamp);

    progress = calculate_progress(num_parts, num_parts_downloaded, num_parts_failed);
    if (progress == 100) {
        progress_str = _("Done");
    } else {
        progress_str = NULL;
    }

    gtk_tree_store_append(store, &iter, &parent);
    gtk_tree_store_set(store, &iter, FIELD_SUBJECT, subject, FIELD_POSTER, poster, FIELD_STAMP_STR, stamp_str, FIELD_FILESIZE_STR, size, FIELD_FILESIZE, file_size, FIELD_NUM_PARTS, num_parts, FIELD_PARTS_DONE, num_parts_downloaded, FIELD_PROGRESS, progress, FIELD_DECODE_PROGRESS_STR, progress_str, -1);

    g_free(size);
    g_free(stamp_str);

    gtk_tree_model_get(GTK_TREE_MODEL(store), &parent, FIELD_NUM_PARTS, &num_parts_total, FIELD_PARTS_DONE, &num_parts_downloaded_total, FIELD_PARTS_FAILED, &num_parts_failed_total, -1);

    num_parts_total += num_parts;
    num_parts_downloaded_total += num_parts_downloaded;
    progress = calculate_progress(num_parts_total, num_parts_downloaded_total, num_parts_failed_total);

    gtk_tree_store_set(GTK_TREE_STORE(store), &parent, FIELD_NUM_PARTS, num_parts_total, FIELD_PARTS_DONE, num_parts_downloaded_total, FIELD_PARTS_FAILED, num_parts_failed_total, FIELD_PROGRESS, progress, -1);
}

static void
foreach_group_func (const char *collection_name, const char *subject, const char *group, gpointer data)
{
    GtkTreeStore *store = GTK_TREE_STORE(data);
    GtkTreeIter iter, parent;
    char *old_groups;
    char *new_groups;

    g_return_if_fail(find_iter_by_collection_name(GTK_TREE_MODEL(store), collection_name, &parent));
    g_return_if_fail(find_iter_by_subject(GTK_TREE_MODEL(store), parent, subject, &iter));

    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, FIELD_GROUPS, &old_groups, -1);
    if (old_groups && strlen(old_groups) > 0) {
        new_groups = g_strdup_printf("%s, %s", old_groups, group);
    } else {
        new_groups = g_strdup(group);
    }

    gtk_tree_store_set(store, &iter, FIELD_GROUPS, new_groups, -1);

    g_free(new_groups);
    g_free(old_groups);
}

static void
on_tree_view_select (GtkTreeSelection *selection, gpointer user_data)
{
    GtkWidget *btnRemoveItemFromQueue = get_widget("btnRemoveItemFromQueue");

    if (gtk_tree_selection_count_selected_rows(selection) == 0) {
        // No items selected, disable the remove button
        gtk_widget_set_sensitive(btnRemoveItemFromQueue, FALSE);
        return;
    }

    // Enable the remove button
    gtk_widget_set_sensitive(btnRemoveItemFromQueue, TRUE);
}

void
on_btnPauseQueue_clicked(GtkWidget *caller, gpointer data)
{
    SchedularState active_state = nntpgrab_schedular_get_state(core);

    if (status_flag == SCHEDULAR_STATE_STOPPED && active_state == SCHEDULAR_STATE_STOPPING) {
        GtkWidget *windowMain;

        windowMain = get_widget("windowMain");

        dialog_show(windowMain, _("The schedular is currently being stopped"), GTK_MESSAGE_INFO, GTK_BUTTONS_OK);

        g_signal_handlers_block_by_func (G_OBJECT (caller), G_CALLBACK (on_btnPauseQueue_clicked), data);

        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(caller), TRUE);

        g_signal_handlers_unblock_by_func (G_OBJECT (caller), G_CALLBACK (on_btnPauseQueue_clicked), data);

        return;
    }

    if (status_flag == SCHEDULAR_STATE_RUNNING) {
        status_flag = SCHEDULAR_STATE_STOPPED;
        nntpgrab_schedular_stop(core, FALSE);
    } else {
        status_flag = SCHEDULAR_STATE_RUNNING;
        nntpgrab_schedular_start(core);
    }
}

static void
schedular_state_changed_cb(NNTPGrabCore *core, SchedularState state, const char *reason)
{
    GtkWidget *btnPauseQueue;

    if (state == SCHEDULAR_STATE_STOPPING) {
        status_flag = SCHEDULAR_STATE_STOPPED;
    } else {
        status_flag = state;
    }

    if (state == SCHEDULAR_STATE_STOPPING && reason != NULL) {
        char *msg;
        GtkWidget *windowMain;

        windowMain = get_widget("windowMain");
        msg = g_strdup_printf(_("The schedular has been paused due to an error:\n%s"), reason);
        dialog_show(windowMain, msg, GTK_MESSAGE_WARNING, GTK_BUTTONS_OK);
    }

    btnPauseQueue = get_widget("btnPauseQueue");
    g_signal_handlers_block_by_func (G_OBJECT (btnPauseQueue), G_CALLBACK (on_btnPauseQueue_clicked), NULL);

    if (state == SCHEDULAR_STATE_RUNNING) {
        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(btnPauseQueue), FALSE);
    } else {
        gtk_toggle_tool_button_set_active(GTK_TOGGLE_TOOL_BUTTON(btnPauseQueue), TRUE);
    }

    g_signal_handlers_unblock_by_func (G_OBJECT (btnPauseQueue), G_CALLBACK (on_btnPauseQueue_clicked), NULL);

}

void
queue_initialize()
{
    GtkWidget *treeDownloadQueue;
    GtkTreeStore *store;
    GtkCellRenderer *cellRenderer;
    GtkTreeSelection *selection;

    nntpgrab_schedular_initialize(core);

    treeDownloadQueue = get_widget("treeDownloadQueue");

    cellRenderer = gtk_cell_renderer_progress_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeDownloadQueue), -1, _("Progress"), cellRenderer, "value", FIELD_PROGRESS, "text", FIELD_DECODE_PROGRESS_STR, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeDownloadQueue), -1, _("Subject"), cellRenderer, "text", FIELD_SUBJECT, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeDownloadQueue), -1, _("Poster"), cellRenderer, "text", FIELD_POSTER, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeDownloadQueue), -1, _("Group"), cellRenderer, "text", FIELD_GROUPS, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeDownloadQueue), -1, _("File size"), cellRenderer, "text", FIELD_FILESIZE_STR, NULL);

    cellRenderer = gtk_cell_renderer_text_new();
    gtk_tree_view_insert_column_with_attributes(GTK_TREE_VIEW(treeDownloadQueue), -1, _("Stamp"), cellRenderer, "text", FIELD_STAMP_STR, NULL);

    store = gtk_tree_store_new( 12,
                                G_TYPE_STRING /* FIELD_COLLECTION_NAME */,
                                G_TYPE_STRING /* FIELD_SUBJECT */,
                                G_TYPE_STRING /* FIELD_POSTER */,
                                G_TYPE_STRING /* FIELD_GROUPS */,
                                G_TYPE_STRING /* FIELD_FILESIZE_STR */,
                                G_TYPE_STRING /* FIELD_STAMP_STR */,
                                G_TYPE_INT /* FIELD_PROGRESS */,
                                G_TYPE_UINT64 /* FIELD_FILESIZE */,
                                G_TYPE_INT /* FIELD_NUM_PARTS */,
                                G_TYPE_INT /* FIELD_PARTS_DONE */,
                                G_TYPE_INT /* FIELD_PARTS_FAILED */,
                                G_TYPE_STRING /* FIELD_DECODE_PROGRESS_STR */);

    gtk_tree_view_set_model(GTK_TREE_VIEW(treeDownloadQueue), GTK_TREE_MODEL(store));

    selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(treeDownloadQueue));
    gtk_tree_selection_set_mode(GTK_TREE_SELECTION(selection), GTK_SELECTION_MULTIPLE);

    g_signal_connect(selection, "changed", G_CALLBACK(on_tree_view_select), NULL);

    nntpgrab_schedular_foreach_task(core, foreach_collection_func, foreach_file_func, foreach_group_func, store);

    g_signal_connect(core, "collection_added", G_CALLBACK(collection_added), NULL);
    g_signal_connect(core, "collection_removed", G_CALLBACK(collection_removed), NULL);
    g_signal_connect(core, "file_added", G_CALLBACK(file_added), NULL);
    g_signal_connect(core, "file_removed", G_CALLBACK(file_removed), NULL);
    g_signal_connect(core, "file_restarted", G_CALLBACK(file_restarted), NULL);
    g_signal_connect(core, "file_download_done", G_CALLBACK(file_download_done), NULL);
    g_signal_connect(core, "file_decode_begin", G_CALLBACK(file_decode_begin), NULL);
    g_signal_connect(core, "file_decode_done", G_CALLBACK(file_decode_done), NULL);
    g_signal_connect(core, "part_done", G_CALLBACK(part_done), NULL);
    g_signal_connect(core, "part_failed", G_CALLBACK(part_failed), NULL);
    g_signal_connect(core, "schedular_state_changed", G_CALLBACK(schedular_state_changed_cb), NULL);

    nntpgrab_schedular_start(core);
}
