/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <glib-object.h>
#include <glib/gi18n.h>
#ifdef WIN32
#include <winsock2.h>
#endif
#include "nntpgrab_plugin_schedular.h"

typedef struct SchedularPlugin SchedularPlugin;

#define SCHEDULAR_PLUGIN_TYPE_OBJECT       (schedular_plugin_get_type ())
#define SCHEDULAR_PLUGIN(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), SCHEDULAR_PLUGIN_TYPE_OBJECT, SchedularPlugin))
#define SCHEDULAR_PLUGIN_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), SCHEDULAR_PLUGIN_TYPE_OBJECT, SchedularPluginClass))
#define IS_SCHEDULAR_PLUGIN(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), SCHEDULAR_PLUGIN_TYPE_OBJECT))
#define IS_SCHEDULAR_PLUGIN_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), SCHEDULAR_PLUGIN_TYPE_OBJECT))
#define SCHEDULAR_PLUGIN_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), SCHEDULAR_PLUGIN_TYPE_OBJECT, SchedularPluginClass))

typedef struct SchedularPluginClass SchedularPluginClass;

static PluginSchedularImportedFuncs imported_funcs;
static SchedularPlugin *schedular;

static GType schedular_plugin_get_type (void);

// queue.c
gboolean     load_download_queue(GList **ret, char **errmsg);
gboolean     save_download_queue(GList *queue, char **errmsg);
void         free_download_queue(GList *queue);

static gboolean schedular_save_queue(char **errmsg);

struct SchedularPlugin
{
    GObject parent;

    // The rwlock is used for settings properties used by the schedular thread
    GStaticRWLock rwlock;
    GList *queue;           // Pointers to NNTPCollection*
    GList *servers;         // Pointers to ConfigServer*

    // The mutex is used for the creation and destruction of the schedular thread
    GStaticMutex mutex;
    GThread *thread;
    gboolean abort_flag;
};

struct SchedularPluginClass
{
    GObjectClass parent;
};

typedef struct _decoder_data {
    SchedularPlugin *schedular;
    const char *collection_name;
    NNTPFile *file;
} DecoderData;

G_DEFINE_TYPE(SchedularPlugin, schedular_plugin, G_TYPE_OBJECT);

gboolean
nntpgrab_plugin_schedular_initialize(PluginSchedularImportedFuncs funcs, char **errmsg)
{
#ifdef WIN32
    // Initialise Winsock
   WORD wVersionRequested;
   WSADATA wsaData;
   int err;

   wVersionRequested = MAKEWORD( 2, 2 );

   err = WSAStartup( wVersionRequested, &wsaData );
   if ( err != 0 ) {
       /* Tell the user that we could not find a usable */
       /* WinSock DLL.                                  */
       if (errmsg) {
           *errmsg = g_strdup(_("Winsock could not be initialised"));
       }

       return FALSE;
   }

   /* Confirm that the WinSock DLL supports 2.2.*/
   /* Note that if the DLL supports versions greater    */
   /* than 2.2 in addition to 2.2, it will still return */
   /* 2.2 in wVersion since that is the version we      */
   /* requested.                                        */

   if ( LOBYTE( wsaData.wVersion ) != 2 ||
        HIBYTE( wsaData.wVersion ) != 2 ) {
        /* Tell the user that we could not find a usable */
        /* WinSock DLL.                                  */
        WSACleanup( );

        if (errmsg) {
            *errmsg = g_strdup(_("Winsock 2.2 or higher is required for this program"));
        }

        return FALSE;
     }
#endif

    imported_funcs = funcs;

    schedular = g_object_new(SCHEDULAR_PLUGIN_TYPE_OBJECT, NULL);

    if (!load_download_queue(&schedular->queue, errmsg)) {
        g_object_unref(schedular);

        return FALSE;
    }

    return TRUE;
}

void
nntpgrab_plugin_schedular_destroy(void)
{
    g_object_unref(schedular);
}

int
nntpgrab_plugin_schedular_get_version(void)
{
    return NNTPGRAB_PLUGIN_API_VERSION;
}

static void
schedular_plugin_init (SchedularPlugin *obj)
{
    g_static_rw_lock_init(&obj->rwlock);
    g_static_mutex_init(&obj->mutex);

    obj->thread = NULL;
    obj->abort_flag = FALSE;
    obj->queue = NULL;
}

static void
schedular_plugin_finalize (GObject *obj)
{
    SchedularPlugin *schedular = SCHEDULAR_PLUGIN(obj);

    g_static_mutex_lock(&schedular->mutex);

    if (schedular->thread) {
        g_static_rw_lock_writer_lock(&schedular->rwlock);
        schedular->abort_flag = TRUE;
        g_static_rw_lock_writer_unlock(&schedular->rwlock);

        g_thread_join(schedular->thread);
    }

    g_static_mutex_unlock(&schedular->mutex);

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    save_download_queue(schedular->queue, NULL);        // TODO: Error message not caught

    free_download_queue(schedular->queue);

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    g_static_rw_lock_free(&schedular->rwlock);
    g_static_mutex_free(&schedular->mutex);
}

static void
schedular_plugin_class_init (SchedularPluginClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = schedular_plugin_finalize;
}

static gboolean
get_next_part_to_download(SchedularPlugin *schedular, ConfigServer *server, NNTPCollection **collection, NNTPFile **file, NNTPPart **part, GThreadPool *poolDecoder)
{
    GList *list;

    g_static_rw_lock_reader_lock(&schedular->rwlock);

    list = schedular->queue;
    while (list) {
        GList *list2;

        *collection = (NNTPCollection *) list->data;

        // Is this collection being removed ?
        if ((*collection)->stop_flag) {
            list = g_list_next(list);
            continue;
        }

        list2 = (*collection)->files;
        while (list2) {
            GList *list3;
            DecoderData *decoder_data;
            gboolean complete_flag;

            *file = (NNTPFile *) list2->data;

            // Is this file already downloaded?
            if ((*file)->file_is_downloaded) {
                list2 = g_list_next(list2);
                continue;
            }

            if ((*file)->now_decoding) {
                list2 = g_list_next(list2);
                continue;
            }

            // Is this file being removed?
            if ((*file)->stop_flag) {
                list2 = g_list_next(list2);
                continue;
            }

            complete_flag = TRUE;
            list3 = (*file)->parts;
            while (list3) {
                *part = (NNTPPart *) list3->data;

                g_static_rw_lock_reader_unlock(&schedular->rwlock);
                g_static_rw_lock_writer_lock(&schedular->rwlock);
                if ((*part)->now_downloading) {
                    complete_flag = FALSE;
                }

                if (!(*part)->downloaded) {
                    complete_flag = FALSE;

                    if (!(*part)->now_downloading) {
                        (*part)->now_downloading = TRUE;
                        (*file)->file_is_downloading = TRUE;
                        g_static_rw_lock_writer_unlock(&schedular->rwlock);

                        return TRUE;
                    }
                }
                g_static_rw_lock_writer_unlock(&schedular->rwlock);
                g_static_rw_lock_reader_lock(&schedular->rwlock);

                list3 = g_list_next(list3);
            }

            // All the parts are already downloaded?
            g_static_rw_lock_reader_unlock(&schedular->rwlock);
            g_static_rw_lock_writer_lock(&schedular->rwlock);

            if (complete_flag && !(*file)->now_decoding) {
                (*file)->now_decoding = TRUE;

                imported_funcs.notify_download_done((*collection)->collection_name, (*file)->subject);

                decoder_data = g_malloc(sizeof(DecoderData));
                decoder_data->collection_name = (*collection)->collection_name;
                decoder_data->file = (*file);
                decoder_data->schedular = schedular;

                g_thread_pool_push(poolDecoder, decoder_data, NULL);
            }

            g_static_rw_lock_writer_unlock(&schedular->rwlock);
            g_static_rw_lock_reader_lock(&schedular->rwlock);

            list2 = g_list_next(list2);
        }

        list = g_list_next(list);
    }

    g_static_rw_lock_reader_unlock(&schedular->rwlock);

    *collection = NULL;
    *file = NULL;
    *part = NULL;

    return FALSE;
}

static void
update_part_download_status(NNTPFile *file, NNTPPart *part, gboolean success)
{
    GList *list;

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    part->now_downloading = FALSE;
    part->downloaded = success;
    file->file_is_downloading = FALSE;

    // Are other threads downloading parts of this file ?
    list = file->parts;
    while (list) {
        NNTPPart *part2 = list->data;

        if (part2->now_downloading) {
            file->file_is_downloading = TRUE;
            g_static_rw_lock_writer_unlock(&schedular->rwlock);
            return;
        }

        list = g_list_next(list);
    }

    g_static_rw_lock_writer_unlock(&schedular->rwlock);
}

static gpointer
stop_schedular_func(gpointer data)
{
    nntpgrab_plugin_schedular_stop((const char *) data);

    g_free(data);

    return NULL;
}

static void
stop_schedular_from_seperate_thread(const char *reason)
{
    // The schedular needs to be stopped from a seperate thread in order to prevent deadlocks
    g_thread_create(stop_schedular_func, g_strdup(reason), FALSE, NULL);
}

static void
decoder_thread_func(gpointer data, gpointer user_data)
{
    DecoderData *decoder_data = (DecoderData *) data;
    gboolean do_emit;
    int saved_errno = 0;

    imported_funcs.notify_decode_begin(decoder_data->collection_name, decoder_data->file->subject);

    // Perform the actual decoding
    if (!imported_funcs.decode_file(decoder_data->collection_name, decoder_data->file, &saved_errno)) {
        if (saved_errno == ENOSPC) {
            // No free space available, stop the schedular
            stop_schedular_from_seperate_thread(strerror(saved_errno));

            // Sleep for 1 second to prevent other threads from retrying this file
            g_usleep(G_USEC_PER_SEC * 1);

            g_static_rw_lock_writer_lock(&decoder_data->schedular->rwlock);
            decoder_data->file->now_decoding = FALSE;
            g_static_rw_lock_writer_unlock(&decoder_data->schedular->rwlock);

            g_free(decoder_data);

            return;
        } else if (saved_errno == ENOENT) {
            // No parts available, ignore
        } else {
            // Unknown decode error occured, stop the schedular
            char *msg = g_strdup_printf("Unknown decode error occured, %s (%i)", strerror(saved_errno), saved_errno);
            stop_schedular_from_seperate_thread(msg);
            g_free(msg);

            // Sleep for 1 second to prevent other threads from retrying this file
            g_usleep(G_USEC_PER_SEC * 1);

            g_static_rw_lock_writer_lock(&decoder_data->schedular->rwlock);
            decoder_data->file->now_decoding = FALSE;
            g_static_rw_lock_writer_unlock(&decoder_data->schedular->rwlock);

            g_free(decoder_data);

            return;
        }
    }

    // Mark the file as completed
    g_static_rw_lock_writer_lock(&decoder_data->schedular->rwlock);
    // Was the task restarted while the decoding was busy?
    if (decoder_data->file->now_decoding) {
        decoder_data->file->file_is_downloaded = TRUE;
        decoder_data->file->now_decoding = FALSE;
        do_emit = TRUE;
    } else {
        do_emit = FALSE;
    }
    g_static_rw_lock_writer_unlock(&decoder_data->schedular->rwlock);

    if (do_emit) {
        imported_funcs.notify_decode_done(decoder_data->collection_name, decoder_data->file->subject);
    }

    g_free(decoder_data);

    schedular_save_queue(NULL);
}

typedef struct _download_data {
    ConfigServer *server;
    GThreadPool *poolDecoder;
} DownloadData;

static void
download_thread_func(gpointer data, gpointer user_data)
{
    DownloadData *download_data = (DownloadData *) data;
    SchedularPlugin *schedular = SCHEDULAR_PLUGIN(user_data);
    NNTPPart *part;
    NNTPFile *file;
    NNTPCollection *collection;
    NNTPGrabErrCode err;
    gboolean abort_flag = FALSE;
    gpointer nntp_connection = NULL;
    gint idle_timer = 0;
    char *reason;

    do {
        g_static_rw_lock_reader_lock(&schedular->rwlock);
        abort_flag = schedular->abort_flag;
        g_static_rw_lock_reader_unlock(&schedular->rwlock);

        // Have we got anything to download ?
        if (!(get_next_part_to_download(schedular, download_data->server, &collection, &file, &part, download_data->poolDecoder))) {
            idle_timer++;
            if (idle_timer == 10) {
                // We have waited for 10 seconds with no new tasks. Kill the connection
                if (nntp_connection) {
                    imported_funcs.nntp_connection_destroy(nntp_connection);
                    nntp_connection = NULL;
                }
            }

            g_usleep(G_USEC_PER_SEC * 1);
            continue;
        }

        idle_timer = 0;

        // Perform the download
        if (!nntp_connection) {
            err = NNTP_ERROR_NONE;
            nntp_connection = imported_funcs.nntp_connection_initialize(download_data->server->servername, &err);
        }

        if (!nntp_connection) {
            // Connection FAILED. Put the part back in the queue and sleep for a few seconds to avoid flooding
            update_part_download_status(file, part, FALSE);

            if (err == NNTP_ERROR_HOST_NOT_FOUND) {
                // Configuration error detected, kill this thread
                break;
            }

            g_usleep(G_USEC_PER_SEC * 5);

            continue;
        }

        // Download the part.
        err = imported_funcs.nntp_connection_get_part(nntp_connection, collection->collection_name, file, part, &reason);
        switch (err) {
            case NNTP_ERROR_NONE:
                update_part_download_status(file, part, TRUE);

                break;

            case NNTP_ERROR_PART_NOT_AVAILABLE:
                // FIXME: ignore for now, needs to be fixed when multiserver support arrives
                update_part_download_status(file, part, TRUE);

                break;

            case NNTP_ERROR_UNABLE_TO_SAVE_PART:
                update_part_download_status(file, part, TRUE);

                stop_schedular_from_seperate_thread(reason);

                if (reason) {
                    g_free(reason);
                }

                // Prevent to start a new download task
                g_usleep(G_USEC_PER_SEC / 10);

                break;

            default:
                // Connection error. Put the part back in the queue and
                // sleep for a few seconds to avoid flooding

                update_part_download_status(file, part, FALSE);

                // Kill the connection
                imported_funcs.nntp_connection_destroy(nntp_connection);
                nntp_connection = NULL;

                g_usleep(G_USEC_PER_SEC * 5);

                break;
        };

#if 0
        // Are all the parts of this file now downloaded ?
        if (is_file_downloaded(schedular, file)) {
            DecoderData *decoder_data;

            imported_funcs.notify_download_done(collection->collection_name, file->subject);

            decoder_data = g_malloc(sizeof(DecoderData));
            decoder_data->collection_name = collection->collection_name;
            decoder_data->file = file;
            decoder_data->schedular = schedular;

            g_thread_pool_push(download_data->poolDecoder, decoder_data, NULL);
        }
#endif

        g_static_rw_lock_reader_lock(&schedular->rwlock);
        abort_flag = schedular->abort_flag;
        g_static_rw_lock_reader_unlock(&schedular->rwlock);
    } while (!abort_flag);

    if (nntp_connection) {
        imported_funcs.nntp_connection_destroy(nntp_connection);
    }

    g_free(download_data);
}

static gpointer
schedular_thread_func(gpointer data)
{
    SchedularPlugin *schedular = SCHEDULAR_PLUGIN(data);
    GThreadPool *pool[10];
    GThreadPool *poolDecoder;
    GList *list;
    int i = 0;
    gboolean abort_flag;

    // Initialize a thread pool for all the configured servers
    g_assert(g_list_length(schedular->servers) < 10);
    memset(pool, 0, sizeof(pool));

    poolDecoder = g_thread_pool_new(decoder_thread_func, schedular, 1, FALSE, NULL);

    list = schedular->servers;
    while (list) {
        ConfigServer *server = list->data;
        int j;

        g_assert(i < 10);

        pool[i] = g_thread_pool_new(download_thread_func, schedular, server->max_threads, FALSE, NULL);

        g_assert(pool[i]);

        // Launch all threads in the pool
        for (j = 0; j < server->max_threads; j++) {
            DownloadData *download_data = g_malloc(sizeof(DownloadData));
            download_data->server = server;
            download_data->poolDecoder = poolDecoder;
            g_thread_pool_push(pool[i], download_data, NULL);
        }

        i++;

        list = g_list_next(list);
    }

    abort_flag = FALSE;
    while (!abort_flag) {
        g_static_rw_lock_reader_lock(&schedular->rwlock);
        abort_flag = schedular->abort_flag;
        g_static_rw_lock_reader_unlock(&schedular->rwlock);

        g_usleep(G_USEC_PER_SEC * 1);
    }

    for (i = 0; i < 10; i++) {
        if (pool[i]) {
            g_thread_pool_free(pool[i], TRUE, TRUE);
        }
    }

    g_thread_pool_free(poolDecoder, TRUE, TRUE);

    return NULL;
}

gboolean
nntpgrab_plugin_schedular_start(void)
{
    GError *error = NULL;
    GList *list;
    GList *servers;

    g_assert(schedular != NULL);

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    if (schedular->thread != NULL) {
        g_static_rw_lock_writer_unlock(&schedular->rwlock);
        return FALSE;
    }

    // Make a copy of the servers list
    g_assert(schedular->servers == NULL);

    servers = imported_funcs.config_get_avail_servers();
    list = servers;
    while (list) {
        char *servername = list->data;
        ConfigServer *server = imported_funcs.config_get_server_info(servername);

        g_assert(server);

        schedular->servers = g_list_append(schedular->servers, server);

        g_free(servername);

        list = g_list_next(list);
    }

    g_list_free(servers);

    // Reset the abort flag
    schedular->abort_flag = FALSE;

    // Initialize and start the real schedular thread
    if (!(schedular->thread = g_thread_create(schedular_thread_func, schedular, TRUE, &error))) {
        imported_funcs.emit_fatal_error(error->message);
        g_error_free(error);
    }

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    imported_funcs.notify_schedular_state_changed(SCHEDULAR_STATE_RUNNING, NULL);

    return TRUE;
}

gboolean
nntpgrab_plugin_schedular_stop(const char *reason)
{
    GList *list;

    g_assert(schedular != NULL);

    g_static_mutex_lock(&schedular->mutex);

    if (schedular->thread == NULL) {
        g_static_mutex_unlock(&schedular->mutex);
        return FALSE;
    }

    imported_funcs.notify_schedular_state_changed(SCHEDULAR_STATE_STOPPING, reason);

    // Abort the schedular thread
    g_static_rw_lock_writer_lock(&schedular->rwlock);
    schedular->abort_flag = TRUE;
    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    g_thread_join(schedular->thread);
    schedular->thread = NULL;

    // Clear all the configured usenet servers
    g_static_rw_lock_writer_lock(&schedular->rwlock);
    list = schedular->servers;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }
    g_list_free(schedular->servers);
    schedular->servers = NULL;
    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    g_static_mutex_unlock(&schedular->mutex);

    imported_funcs.notify_schedular_state_changed(SCHEDULAR_STATE_STOPPED, NULL);

    return TRUE;
}

SchedularState
nntpgrab_plugin_schedular_get_state(void)
{
    SchedularState state;

    g_static_rw_lock_reader_lock(&schedular->rwlock);
    if (schedular->thread && schedular->abort_flag) {
        state = SCHEDULAR_STATE_STOPPING;
    } else if (schedular->thread && !schedular->abort_flag) {
        state = SCHEDULAR_STATE_RUNNING;
    } else {
        state = SCHEDULAR_STATE_STOPPED;
    }
    g_static_rw_lock_reader_unlock(&schedular->rwlock);

    return state;
}

static gboolean
test_is_subject_already_in_collection(NNTPCollection *collection, const char *subject)
{
    GList *list;

    list = collection->files;
    while (list) {
        NNTPFile *file = (NNTPFile *) list->data;

        if (!strcmp(file->subject, subject)) {
            return TRUE;
        }

        list = g_list_next(list);
    }

    return FALSE;
}

gboolean
nntpgrab_plugin_schedular_add_task_to_queue(const char *collection_name, NNTPFile *file, char **errmsg)
{
    GList *list;
    NNTPCollection *collection;

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    // Find out if the collection_name already exists in the download queue
    list = schedular->queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        if (!strcmp(collection->collection_name, collection_name)) {
            // Collection found!

            // Does the subject already exists in this collection ?
            if (test_is_subject_already_in_collection(collection, file->subject)) {
                if (errmsg) {
                    *errmsg = g_strdup_printf(_("There already is a file with subject '%s' part of collection '%s'"), file->subject, collection_name);
                }

                g_static_rw_lock_writer_unlock(&schedular->rwlock);

                return FALSE;
            }

            // Add the file to this collection
            collection->files = g_list_append(collection->files, file);
            collection->total_size += file->file_size;
            file->position = g_list_index(collection->files, file) + 1;

            imported_funcs.task_added(collection_name, file);

            g_static_rw_lock_writer_unlock(&schedular->rwlock);

            return TRUE;
        }

        list = g_list_next(list);
    }

    // Collection doesn't exist yet, add a new one
    collection = g_malloc0(sizeof(NNTPCollection));
    strncpy((char *) collection->collection_name, collection_name, sizeof(collection->collection_name));
    schedular->queue = g_list_append(schedular->queue, collection);

    imported_funcs.collection_added(collection_name);

    // Finally add the task to the collection
    collection->files = g_list_append(collection->files, file);
    collection->total_size = file->file_size;
    file->position = 1;

    imported_funcs.task_added(collection_name, file);

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    return TRUE;
}

static void
free_file(NNTPFile *file)
{
    GList *list;

    g_assert(file);

    list = file->parts;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }

    list = file->groups;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }

    g_free(file);
}

gboolean
nntpgrab_plugin_schedular_del_task_from_queue(const char *collection_name, const char *subject, char **errmsg)
{
    GList *list;
    NNTPCollection *collection;

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    // Find out if the collection_name exists in the download queue
    list = schedular->queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        if (!strcmp(collection->collection_name, collection_name)) {
            // Collection found!
            GList *list2;
            gboolean delete_flag = FALSE;

            // Do we have a subject?
            if (!subject) {
                // Delete the whole collection
                delete_flag = TRUE;

                // Mark the whole collection as being removed
                collection->stop_flag = TRUE;
            }

            // Try to find the subject
            list2 = collection->files;
            while (list2) {
                NNTPFile *file = (NNTPFile *) list2->data;

                list2 = g_list_next(list2);

                if (delete_flag || !strcmp(file->subject, subject)) {
                    // Bingo!
                    // Is this file being downloaded ?
                    while (file->file_is_downloading || file->now_decoding) {
                        // Mark this file as being removed
                        file->stop_flag = TRUE;

                        // Release the lock and wait for the download to complete
                        g_static_rw_lock_writer_unlock(&schedular->rwlock);

                        g_usleep(G_USEC_PER_SEC / 10);

                        g_static_rw_lock_writer_lock(&schedular->rwlock);
                    }

                    // Update the size of the collection
                    collection->total_size -= file->file_size;

                    // Remove it from the list
                    collection->files = g_list_remove(collection->files, file);

                    // Restart from the beginning of the list as we have removed an item
                    list2 = collection->files;

                    imported_funcs.task_removed(collection_name, file->subject);

                    free_file(file);

                    // Does the collection still got any items in it ?
                    if (g_list_length(collection->files) == 0) {
                        // Yes, remove the entire collection
                        schedular->queue = g_list_remove(schedular->queue, collection);

                        imported_funcs.collection_removed(collection_name);

                        g_free(collection);
                    }

                    if (!delete_flag) {
                        g_static_rw_lock_writer_unlock(&schedular->rwlock);
                        return TRUE;
                    }
                }
            }

            if (delete_flag) {
                g_static_rw_lock_writer_unlock(&schedular->rwlock);
                return TRUE;
            }
        }

        list = g_list_next(list);
    }

    if (errmsg) {
        *errmsg = g_strdup_printf(_("Subject '%s' was not found in collection '%s'"), subject, collection_name);
    }

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    return FALSE;
}

gboolean
nntpgrab_plugin_schedular_restart_task(const char *collection_name, const char *subject, char **errmsg)
{
    GList *list;
    NNTPCollection *collection;

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    // Find out if the collection_name already exists in the download queue
    list = schedular->queue;
    while (list) {
        collection = (NNTPCollection *) list->data;

        if (!strcmp(collection->collection_name, collection_name)) {
            // Collection found!
            GList *list2;
            gboolean restart_flag = FALSE;

            // Do we have a subject?
            if (!subject) {
                // Restart the whole collection
                restart_flag = TRUE;
            }

            list2 = collection->files;
            while (list2) {
                NNTPFile *file = list2->data;

                if (restart_flag || !strcmp(file->subject, subject)) {
                    // Subject found!
                    GList *list3;

                    list3 = file->parts;
                    while (list3) {
                        NNTPPart *part = list3->data;

                        part->downloaded = FALSE;
                        part->now_downloading = FALSE;

                        list3 = g_list_next(list3);
                    }

                    file->file_is_downloading = FALSE;
                    file->file_is_downloaded = FALSE;
                    file->now_decoding = FALSE;

                    if (restart_flag) {
                        imported_funcs.task_restarted(collection_name, file->subject);
                    } else {
                        g_static_rw_lock_writer_unlock(&schedular->rwlock);

                        imported_funcs.task_restarted(collection_name, file->subject);

                        return TRUE;
                    }
                }

                list2 = g_list_next(list2);
            }

            if (restart_flag) {
                g_static_rw_lock_writer_unlock(&schedular->rwlock);
                return TRUE;
            }
        }

        list = g_list_next(list);
    }

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    if (errmsg) {
        *errmsg = g_strdup_printf(_("Subject '%s' was not found in collection '%s'"), subject, collection_name);
    }

    return FALSE;
}

gboolean
nntpgrab_plugin_schedular_save_queue(char **errmsg)
{
    return schedular_save_queue(errmsg);
}


void
nntpgrab_plugin_schedular_foreach_task(GFunc foreach_func, gpointer data)
{
    g_assert(schedular != NULL);

    g_static_rw_lock_reader_lock(&schedular->rwlock);

    g_list_foreach(schedular->queue, foreach_func, data);

    g_static_rw_lock_reader_unlock(&schedular->rwlock);
}

static gboolean
schedular_save_queue(char **errmsg)
{
    g_assert(schedular != NULL);

    g_static_rw_lock_writer_lock(&schedular->rwlock);

    if (!save_download_queue(schedular->queue, errmsg)) {
        g_static_rw_lock_writer_unlock(&schedular->rwlock);
        return FALSE;
    }

    g_static_rw_lock_writer_unlock(&schedular->rwlock);

    return TRUE;
}

