/*
    Copyright (C) 2005-2007  Erik van Pienbroek

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include "nntpgrab.h"
#include "nntpgrab_internal.h"
#include "marshalers.h"
#include "configuration.h"
#include "schedular.h"

#define NNTPGRAB_TYPE_CORE              (core_get_type ())
#define NNTPGRAB_CORE(object)           (G_TYPE_CHECK_INSTANCE_CAST ((object), NNTPGRAB_TYPE_CORE, NNTPGrabCore))
#define NNTPGRAB_CORE_CLASS(klass)      (G_TYPE_CHECK_CLASS_CAST ((klass), NNTPGRAB_TYPE_CORE, NNTPGrabCoreClass))
#define IS_NNTPGRAB_CORE(object)        (G_TYPE_CHECK_INSTANCE_TYPE ((object), NNTPGRAB_TYPE_CORE))
#define IS_NNTPGRAB_CORE_CLASS(klass)   (G_TYPE_CHECK_CLASS_TYPE ((klass), NNTPGRAB_TYPE_CORE))
#define NNTPGRAB_CORE_GET_CLASS(obj)    (G_TYPE_INSTANCE_GET_CLASS ((obj), NNTPGRAB_TYPE_CORE, NNTPGrabCoreClass))

struct NNTPGrabCore
{
    GObject parent;

    Configuration *config;
    Schedular *schedular;
    gboolean is_finalizing;
};

static NNTPGrabCore *core = NULL;

struct NNTPGrabCoreClass
{
    GObjectClass parent;
};

enum
{
    CONFIG_CHANGED_SIGNAL,
    FILE_DOWNLOAD_DONE_SIGNAL,
    FILE_DECODE_BEGIN_SIGNAL,
    FILE_DECODE_DONE_SIGNAL,
    PART_DOWNLOAD_START_SIGNAL,
    PART_DONE_SIGNAL,
    PART_FAILED_SIGNAL,
    PART_PROGRESS_UPDATE_SIGNAL,
    TRAFFIC_MONITOR_UPDATE_SIGNAL,
    COLLECTION_ADDED_SIGNAL,
    COLLECTION_REMOVED_SIGNAL,
    FILE_ADDED_SIGNAL,
    FILE_REMOVED_SIGNAL,
    FILE_RESTARTED_SIGNAL,
    CONNECTION_CONNECTING_SIGNAL,
    CONNECTION_CONNECTED_SIGNAL,
    CONNECTION_DISCONNECT_SIGNAL,
    SCHEDULAR_STATE_CHANGED_SIGNAL,
    FATAL_ERROR_SIGNAL,
    WARNING_MESSAGE_SIGNAL,
    DEBUG_MESSAGE_SIGNAL,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(NNTPGrabCore, core, G_TYPE_OBJECT)

static void
core_init (NNTPGrabCore *obj)
{
    obj->is_finalizing = FALSE;
}

static void
core_finalize (GObject *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    core->is_finalizing = TRUE;

    if (core->config) {
        configuration_destroy(core->config);
    }

    if (core->schedular) {
        schedular_destroy(core->schedular);
    }
}

static void
core_class_init (NNTPGrabCoreClass *klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

    gobject_class->finalize = core_finalize;

    signals[CONFIG_CHANGED_SIGNAL] =       g_signal_new("config_changed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__VOID,
                                                        G_TYPE_NONE, 0);

    signals[FILE_DOWNLOAD_DONE_SIGNAL] =  g_signal_new ("file_download_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[FILE_DECODE_BEGIN_SIGNAL] =   g_signal_new ("file_decode_begin",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[FILE_DECODE_DONE_SIGNAL] =    g_signal_new ("file_decode_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[PART_DOWNLOAD_START_SIGNAL] = g_signal_new ("part_download_start",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PART_DONE_SIGNAL] =           g_signal_new ("part_done",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PART_FAILED_SIGNAL] =         g_signal_new ("part_failed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_INT_STRING_STRING_INT,
                                                        G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);

    signals[PART_PROGRESS_UPDATE_SIGNAL] = g_signal_new ("part_progress_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__STRING_INT_STRING_INT_INT_INT,
                                                         G_TYPE_NONE, 6, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT);

    signals[TRAFFIC_MONITOR_UPDATE_SIGNAL] = g_signal_new ("traffic_monitor_update",
                                                         G_OBJECT_CLASS_TYPE (klass),
                                                         G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                         0,
                                                         NULL, NULL,
                                                         nntpgrab_marshal_VOID__INT_INT_INT_INT_INT_INT_INT_INT_INT_INT_INT64_DOUBLE,
                                                         G_TYPE_NONE, 12, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT, G_TYPE_INT64, G_TYPE_DOUBLE);

    signals[COLLECTION_ADDED_SIGNAL] =    g_signal_new ("collection_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[COLLECTION_REMOVED_SIGNAL] =  g_signal_new ("collection_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING,
                                                        G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[FILE_ADDED_SIGNAL] =          g_signal_new ("file_added",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING_STRING_UINT64_UINT64_INT_POINTER,
                                                        G_TYPE_NONE, 7, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_UINT64, G_TYPE_UINT64, G_TYPE_INT, G_TYPE_POINTER);

    signals[FILE_REMOVED_SIGNAL] =        g_signal_new ("file_removed",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[FILE_RESTARTED_SIGNAL] =      g_signal_new ("file_restarted",
                                                        G_OBJECT_CLASS_TYPE (klass),
                                                        G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                        0,
                                                        NULL, NULL,
                                                        nntpgrab_marshal_VOID__STRING_STRING,
                                                        G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_STRING);

    signals[CONNECTION_CONNECTING_SIGNAL] =  g_signal_new   ("connection_connecting",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT,
                                                             G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_INT);

    signals[CONNECTION_CONNECTED_SIGNAL] =  g_signal_new    ("connection_connected",
                                                             G_OBJECT_CLASS_TYPE (klass),
                                                             G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                             0,
                                                             NULL, NULL,
                                                             nntpgrab_marshal_VOID__STRING_INT_STRING,
                                                             G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING);

    signals[CONNECTION_DISCONNECT_SIGNAL] = g_signal_new ("connection_disconnect",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING_INT_ENUM_STRING,
                                                          G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT, G_TYPE_STRING);


    signals[SCHEDULAR_STATE_CHANGED_SIGNAL] = g_signal_new ("schedular_state_changed",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__INT_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_INT, G_TYPE_STRING);


    signals[FATAL_ERROR_SIGNAL] =           g_signal_new ("fatal_error",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[WARNING_MESSAGE_SIGNAL] =       g_signal_new ("warning_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__STRING,
                                                          G_TYPE_NONE, 1, G_TYPE_STRING);

    signals[DEBUG_MESSAGE_SIGNAL] =         g_signal_new ("debug_message",
                                                          G_OBJECT_CLASS_TYPE (klass),
                                                          G_SIGNAL_RUN_LAST | G_SIGNAL_DETAILED,
                                                          0,
                                                          NULL, NULL,
                                                          nntpgrab_marshal_VOID__POINTER_STRING,
                                                          G_TYPE_NONE, 2, G_TYPE_POINTER, G_TYPE_STRING);
}

/***************************************************************/
/*                                                             */
/*                   Internal methods                          */
/*                                                             */
/***************************************************************/
static gboolean
do_emit_fatal_error(gpointer data)
{
    char *errmsg = (char *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    g_signal_emit_by_name(core, "fatal_error", errmsg);
    g_free(errmsg);

    return FALSE;
}

void
nntpgrab_core_emit_fatal_error(const char *errmsg)
{
    g_return_if_fail(core != NULL);

    g_idle_add(do_emit_fatal_error, (gpointer) g_strdup(errmsg));

g_print("FATAL ERROR: %s\n", errmsg);
}

static gboolean
do_emit_warning(gpointer data)
{
    char *errmsg = (char *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    g_signal_emit_by_name(core, "warning_message", errmsg);
    g_free(errmsg);

    return FALSE;
}

void
nntpgrab_core_emit_warning(const char *errmsg)
{
    g_return_if_fail(core != NULL);

    g_idle_add(do_emit_warning, (gpointer) g_strdup(errmsg));

g_print("WARNING: %s\n", errmsg);
}

struct _do_emit_connecting_data {
    char *servername;
    int conn_id;
};

static gboolean
do_emit_connecting(gpointer data)
{
    struct _do_emit_connecting_data *connecting_data = (struct _do_emit_connecting_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(connecting_data->servername);
        g_free(connecting_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "connection_connecting", connecting_data->servername, connecting_data->conn_id);
    g_free(connecting_data->servername);
    g_free(connecting_data);

    return FALSE;
}

void
nntpgrab_core_emit_connecting(const char *servername, int conn_id)
{
    struct _do_emit_connecting_data *connecting_data = (struct _do_emit_connecting_data *) g_malloc0(sizeof(struct _do_emit_connecting_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    connecting_data->servername = g_strdup(servername);
    connecting_data->conn_id = conn_id;

    g_idle_add(do_emit_connecting, connecting_data);
}

struct _do_emit_connected_data {
    char *servername;
    int conn_id;
    char *welcome_msg;
};

static gboolean
do_emit_connected(gpointer data)
{
    struct _do_emit_connected_data *connected_data = (struct _do_emit_connected_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(connected_data->servername);
        g_free(connected_data->welcome_msg);
        g_free(connected_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "connection_connected", connected_data->servername, connected_data->conn_id, connected_data->welcome_msg);

    g_free(connected_data->servername);
    g_free(connected_data->welcome_msg);
    g_free(connected_data);

    return FALSE;
}

void
nntpgrab_core_emit_connected(const char *servername, int conn_id, const char *welcome_msg)
{
    struct _do_emit_connected_data *connected_data = (struct _do_emit_connected_data *) g_malloc0(sizeof(struct _do_emit_connected_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    connected_data->servername = g_strdup(servername);
    connected_data->conn_id = conn_id;
    connected_data->welcome_msg = g_strdup(welcome_msg);

    g_idle_add(do_emit_connected, connected_data);
}

struct _do_emit_disconnect_data {
    char *servername;
    int conn_id;
    NNTPDisconnectType disconnect_type;
    char *errmsg;
};

static gboolean
do_emit_disconnect(gpointer data)
{
    struct _do_emit_disconnect_data *disconnect_data = (struct _do_emit_disconnect_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(disconnect_data->servername);
        if (disconnect_data->errmsg) {
            g_free(disconnect_data->errmsg);
        }
        g_free(disconnect_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "connection_disconnect", disconnect_data->servername, disconnect_data->conn_id, disconnect_data->disconnect_type, disconnect_data->errmsg);

    g_free(disconnect_data->servername);
    if (disconnect_data->errmsg) {
        g_free(disconnect_data->errmsg);
    }
    g_free(disconnect_data);

    return FALSE;
}

void
nntpgrab_core_emit_disconnect(const char *servername, int conn_id, NNTPDisconnectType disconnect_type, const char *errmsg)
{
    struct _do_emit_disconnect_data *disconnect_data = g_malloc0(sizeof(struct _do_emit_disconnect_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    disconnect_data->servername = g_strdup(servername);
    disconnect_data->conn_id = conn_id;
    disconnect_data->disconnect_type = disconnect_type;
    if (errmsg) {
        disconnect_data->errmsg = g_strdup(errmsg);
    }

    g_idle_add(do_emit_disconnect, disconnect_data);
}

struct _traffic_data {
    int bytes_received[10];
    time_t stamp;
};

static gboolean
do_emit_traffic_monitor_update(gpointer data)
{
    struct _traffic_data *traffic_data = (struct _traffic_data *) data;
    double average;
    int i;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(traffic_data);
        return FALSE;
    }

    // Take the average from the last 5 seconds
    average = 0;
    for (i = 5; i < 10; i++) {
        average += traffic_data->bytes_received[i];
    }
    average /= 5;

    g_signal_emit_by_name(core, "traffic_monitor_update",   traffic_data->bytes_received[0],
                                                            traffic_data->bytes_received[1],
                                                            traffic_data->bytes_received[2],
                                                            traffic_data->bytes_received[3],
                                                            traffic_data->bytes_received[4],
                                                            traffic_data->bytes_received[5],
                                                            traffic_data->bytes_received[6],
                                                            traffic_data->bytes_received[7],
                                                            traffic_data->bytes_received[8],
                                                            traffic_data->bytes_received[9],
                                                            traffic_data->stamp,
                                                            average);

    g_free(traffic_data);

    return FALSE;
}

void
nntpgrab_core_emit_traffic_monitor_update(int bytes_received[10], time_t stamp)
{
    struct _traffic_data *traffic_data = (struct _traffic_data *) g_malloc0(sizeof(struct _traffic_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    memcpy(traffic_data->bytes_received, bytes_received, sizeof(int) * 10);
    traffic_data->stamp = stamp;

    g_idle_add(do_emit_traffic_monitor_update, traffic_data);
}

struct _part_data {
    char *servername;
    int conn_id;
    char *collection_name;
    char *subject;
    int part_num;
};

static gboolean
do_emit_part_download_start(gpointer data)
{
    struct _part_data *part_data = (struct _part_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(part_data->servername);
        g_free(part_data->collection_name);
        g_free(part_data->subject);
        g_free(part_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "part_download_start", part_data->servername, part_data->conn_id, part_data->collection_name, part_data->subject, part_data->part_num);

    g_free(part_data->servername);
    g_free(part_data->collection_name);
    g_free(part_data->subject);
    g_free(part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_download_start(const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    struct _part_data *part_data = (struct _part_data *) g_malloc0(sizeof(struct _part_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    part_data->servername = g_strdup(servername);
    part_data->conn_id = conn_id;
    part_data->collection_name = g_strdup(collection_name);
    part_data->subject = g_strdup(subject);
    part_data->part_num = part_num;

    g_idle_add(do_emit_part_download_start, part_data);
}

static gboolean
do_emit_part_failed(gpointer data)
{
    struct _part_data *part_data = (struct _part_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(part_data->servername);
        g_free(part_data->collection_name);
        g_free(part_data->subject);
        g_free(part_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "part_failed", part_data->servername, part_data->conn_id, part_data->collection_name, part_data->subject, part_data->part_num);

    g_free(part_data->servername);
    g_free(part_data->collection_name);
    g_free(part_data->subject);
    g_free(part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_failed(const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    struct _part_data *part_data = (struct _part_data *) g_malloc0(sizeof(struct _part_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    part_data->servername = g_strdup(servername);
    part_data->conn_id = conn_id;
    part_data->collection_name = g_strdup(collection_name);
    part_data->subject = g_strdup(subject);
    part_data->part_num = part_num;

    g_idle_add(do_emit_part_failed, part_data);
}

static gboolean
do_emit_part_done(gpointer data)
{
    struct _part_data *part_data = (struct _part_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(part_data->servername);
        g_free(part_data->collection_name);
        g_free(part_data->subject);
        g_free(part_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "part_done", part_data->servername, part_data->conn_id, part_data->collection_name, part_data->subject, part_data->part_num);

    g_free(part_data->servername);
    g_free(part_data->collection_name);
    g_free(part_data->subject);
    g_free(part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_done(const char *servername, int conn_id, const char *collection_name, const char *subject, int part_num)
{
    struct _part_data *part_data = (struct _part_data *) g_malloc0(sizeof(struct _part_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    part_data->servername = g_strdup(servername);
    part_data->conn_id = conn_id;
    part_data->collection_name = g_strdup(collection_name);
    part_data->subject = g_strdup(subject);
    part_data->part_num = part_num;

    g_idle_add(do_emit_part_done, part_data);
}

struct _part_progress_data {
    char *servername;
    int conn_id;
    char *subject;
    int part_num;
    int bytes_downloaded;
    int bytes_total;
};

static gboolean
do_emit_part_progress_update(gpointer data)
{
    struct _part_progress_data *part_data = (struct _part_progress_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(part_data->servername);
        g_free(part_data->subject);
        g_free(part_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "part_progress_update", part_data->servername, part_data->conn_id, part_data->subject, part_data->part_num, part_data->bytes_downloaded, part_data->bytes_total);

    g_free(part_data->servername);
    g_free(part_data->subject);
    g_free(part_data);

    return FALSE;
}

void
nntpgrab_core_emit_part_progress_update(const char *servername, int conn_id, const char *subject, int part_num, int bytes_downloaded, int bytes_total)
{
    struct _part_progress_data *part_data = (struct _part_progress_data *) g_malloc0(sizeof(struct _part_progress_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    part_data->servername = g_strdup(servername);
    part_data->conn_id = conn_id;
    part_data->subject = g_strdup(subject);
    part_data->part_num = part_num;
    part_data->bytes_downloaded = bytes_downloaded;
    part_data->bytes_total = bytes_total;

    g_idle_add(do_emit_part_progress_update, part_data);
}

struct _file_done_data {
    char *collection_name;
    char *subject;
};

static gboolean
do_emit_file_download_done(gpointer data)
{
    struct _file_done_data *file_data = (struct _file_done_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(file_data->collection_name);
        g_free(file_data->subject);
        g_free(file_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "file_download_done", file_data->collection_name, file_data->subject);

    g_free(file_data->collection_name);
    g_free(file_data->subject);
    g_free(file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_download_done(const char *collection_name, const char *subject)
{
    struct _file_done_data *file_data = (struct _file_done_data *) g_malloc0(sizeof(struct _file_done_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    file_data->collection_name = g_strdup(collection_name);
    file_data->subject = g_strdup(subject);

    g_idle_add(do_emit_file_download_done, file_data);
}

static gboolean
do_emit_file_decode_begin(gpointer data)
{
    struct _file_done_data *file_data = (struct _file_done_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(file_data->collection_name);
        g_free(file_data->subject);
        g_free(file_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "file_decode_begin", file_data->collection_name, file_data->subject);

    g_free(file_data->collection_name);
    g_free(file_data->subject);
    g_free(file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_decode_begin(const char *collection_name, const char *subject)
{
    struct _file_done_data *file_data = (struct _file_done_data *) g_malloc0(sizeof(struct _file_done_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    file_data->collection_name = g_strdup(collection_name);
    file_data->subject = g_strdup(subject);

    g_idle_add(do_emit_file_decode_begin, file_data);
}

static gboolean
do_emit_file_decode_done(gpointer data)
{
    struct _file_done_data *file_data = (struct _file_done_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(file_data->collection_name);
        g_free(file_data->subject);
        g_free(file_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "file_decode_done", file_data->collection_name, file_data->subject);

    g_free(file_data->collection_name);
    g_free(file_data->subject);
    g_free(file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_decode_done(const char *collection_name, const char *subject)
{
    struct _file_done_data *file_data = (struct _file_done_data *) g_malloc0(sizeof(struct _file_done_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    file_data->collection_name = g_strdup(collection_name);
    file_data->subject = g_strdup(subject);

    g_idle_add(do_emit_file_decode_done, file_data);
}

static gboolean
do_emit_collection_added(gpointer data)
{
    char *collection_name = (char *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(collection_name);
        return FALSE;
    }

    g_signal_emit_by_name(core, "collection_added", collection_name);

    g_free(collection_name);

    return FALSE;
}

void
nntpgrab_core_emit_collection_added(const char *collection_name)
{
    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    g_idle_add(do_emit_collection_added, g_strdup(collection_name));
}

static gboolean
do_emit_collection_removed(gpointer data)
{
    char *collection_name = (char *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(collection_name);
        return FALSE;
    }

    g_signal_emit_by_name(core, "collection_removed", collection_name);

    g_free(collection_name);

    return FALSE;
}

void
nntpgrab_core_emit_collection_removed(const char *collection_name)
{
    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    g_idle_add(do_emit_collection_removed, g_strdup(collection_name));
}

struct _file_added_data {
    char *collection_name;
    char *subject;
    char *poster;
    time_t stamp;
    guint64 file_size;
    int num_parts;
    GList *groups;
};

static gboolean
do_emit_file_added(gpointer data)
{
    struct _file_added_data *file_data = (struct _file_added_data *) data;
    GList *list;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        list = file_data->groups;
        while (list) {
            g_free(list->data);
            list = g_list_next(list);
        }

        g_free(file_data->collection_name);
        g_free(file_data->subject);
        g_free(file_data->poster);
        g_list_free(file_data->groups);
        g_free(file_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "file_added", file_data->collection_name, file_data->subject, file_data->poster, file_data->stamp, file_data->file_size, file_data->num_parts, file_data->groups);

    list = file_data->groups;
    while (list) {
        g_free(list->data);
        list = g_list_next(list);
    }

    g_free(file_data->collection_name);
    g_free(file_data->subject);
    g_free(file_data->poster);
    g_list_free(file_data->groups);
    g_free(file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_added(const char *collection_name, const NNTPFile *file)
{
    struct _file_added_data *file_data = (struct _file_added_data *) g_malloc0(sizeof(struct _file_added_data));
    GList *list;

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    file_data->collection_name = g_strdup(collection_name);
    file_data->subject = g_strdup(file->subject);
    file_data->poster = g_strdup(file->poster);
    file_data->stamp = file->stamp;
    file_data->file_size = file->file_size;
    file_data->num_parts = file->numparts;
    file_data->groups = NULL;

    list = file->groups;
    while (list) {
        char *group = (char *) list->data;
        file_data->groups = g_list_append(file_data->groups, g_strdup(group));
        list = g_list_next(list);
    }

    g_idle_add(do_emit_file_added, file_data);
}

static gboolean
do_emit_file_removed(gpointer data)
{
    struct _file_done_data *file_data = (struct _file_done_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(file_data->collection_name);
        g_free(file_data->subject);
        g_free(file_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "file_removed", file_data->collection_name, file_data->subject);

    g_free(file_data->collection_name);
    g_free(file_data->subject);
    g_free(file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_removed(const char *collection_name, const char *subject)
{
    struct _file_done_data *file_data = (struct _file_done_data *) g_malloc0(sizeof(struct _file_done_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    file_data->collection_name = g_strdup(collection_name);
    file_data->subject = g_strdup(subject);

    g_idle_add(do_emit_file_removed, file_data);
}

static gboolean
do_emit_file_restarted(gpointer data)
{
    struct _file_done_data *file_data = (struct _file_done_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(file_data->collection_name);
        g_free(file_data->subject);
        g_free(file_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "file_restarted", file_data->collection_name, file_data->subject);

    g_free(file_data->collection_name);
    g_free(file_data->subject);
    g_free(file_data);

    return FALSE;
}

void
nntpgrab_core_emit_file_restarted(const char *collection_name, const char *subject)
{
    struct _file_done_data *file_data = (struct _file_done_data *) g_malloc0(sizeof(struct _file_done_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    file_data->collection_name = g_strdup(collection_name);
    file_data->subject = g_strdup(subject);

    g_idle_add(do_emit_file_restarted, file_data);

}

struct _schedular_data {
    SchedularState state;
    char *reason;
};

static gboolean
do_emit_schedular_state_changed(gpointer data)
{
    struct _schedular_data *schedular_data = (struct _schedular_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(schedular_data->reason);
        g_free(schedular_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "schedular_state_changed", schedular_data->state, schedular_data->reason);

    g_free(schedular_data->reason);
    g_free(schedular_data);

    return FALSE;
}

void
nntpgrab_core_emit_schedular_state_changed(SchedularState state, const char *reason)
{
    struct _schedular_data *schedular_data = (struct _schedular_data *) g_malloc0(sizeof(struct _schedular_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    schedular_data->state = state;
    schedular_data->reason = g_strdup(reason);

    g_idle_add(do_emit_schedular_state_changed, schedular_data);
}

struct _debug_data {
    GThread *thread;
    char *message;
};

static gboolean
do_emit_debug_message(gpointer data)
{
    struct _debug_data *debug_data = (struct _debug_data *) data;

    g_return_val_if_fail(core != NULL, FALSE);

    if (core->is_finalizing) {
        g_free(debug_data->message);
        g_free(debug_data);
        return FALSE;
    }

    g_signal_emit_by_name(core, "debug_message", debug_data->thread, debug_data->message);

    g_free(debug_data->message);
    g_free(debug_data);

    return FALSE;
}

void
nntpgrab_core_emit_debug_message(const char *message)
{
    struct _debug_data *debug_data = (struct _debug_data *) g_malloc0(sizeof(struct _debug_data));

    g_return_if_fail(core != NULL);

    if (core->is_finalizing) return;

    debug_data->thread = g_thread_self();
    debug_data->message = g_strdup(message);

    g_idle_add(do_emit_debug_message, debug_data);
}

/***************************************************************/
/*                                                             */
/*                    Public methods                           */
/*                                                             */
/***************************************************************/

NNTPGrabCore *
nntpgrab_core_init(int version, char **err, char **warnings)
{
    static gboolean initialized = FALSE;

    if (initialized) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core already initialized"));
        }
        return NULL;
    }

    g_type_init();
    if (!g_thread_supported ()) g_thread_init (NULL);

    // Check if the API version matches
    if (version != NNTPGRAB_API_VERSION) {
        if (err) {
            *err = g_strdup_printf(_("NNTPGrab Core API mismatch (Core API version = %i, frontend's version = %i)"), NNTPGRAB_API_VERSION, version);
        }
        return NULL;
    }

    core = g_object_new(NNTPGRAB_TYPE_CORE, NULL);

    if (!(core->config = configuration_new(err, warnings))) {
        g_object_unref(core);
        return NULL;
    }

    core->schedular = NULL;     // Cannot be initialised here as the 'fatal_error' signal can not be caught at this point yet

    initialized = TRUE;

    return core;
}

void
nntpgrab_core_cleanup(NNTPGrabCore *obj)
{
    g_object_unref(obj);
}

GList *
nntpgrab_config_get_avail_servers(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->config);

    return configuration_get_avail_servers(core->config);
}

void
nntpgrab_config_free_avail_servers(NNTPGrabCore *obj, GList *servers)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->config);

    configuration_free_avail_servers(core->config, servers);
}

gboolean
nntpgrab_config_get_server_info(NNTPGrabCore *obj, const char *servername, ConfigServer *ret)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    ConfigServer *config;

    g_assert(core->config);

    config = configuration_get_server_info(core->config, servername);
    if (config) {
        memcpy(ret, config, sizeof(ConfigServer));
        g_free(config);
        return TRUE;
    } else {
        return FALSE;
    }
}

gboolean
nntpgrab_config_add_server(NNTPGrabCore *obj, ConfigServer new_server)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg;

    g_assert(core->config);

    retval = configuration_add_server(core->config, new_server);

    if (!configuration_save(core->config, &errmsg)) {
        nntpgrab_core_emit_fatal_error(errmsg);
        g_free(errmsg);
    }

    return retval;
}

gboolean
nntpgrab_config_del_server(NNTPGrabCore *obj, const char *servername)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg;

    g_assert(core->config);

    retval = configuration_del_server(core->config, servername);

    if (!configuration_save(core->config, &errmsg)) {
        nntpgrab_core_emit_fatal_error(errmsg);
        g_free(errmsg);
    }

    return retval;
}

gboolean
nntpgrab_config_edit_server(NNTPGrabCore *obj, const char *servername, ConfigServer server)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    gboolean retval;
    char *errmsg;

    g_assert(core->config);

    retval = configuration_edit_server(core->config, servername, server);

    if (!configuration_save(core->config, &errmsg)) {
        nntpgrab_core_emit_fatal_error(errmsg);
        g_free(errmsg);
    }

    return retval;
}

ConfigOpts
nntpgrab_config_get_opts(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->config);

    return configuration_get_opts(core->config);
}

void
nntpgrab_config_set_opts(NNTPGrabCore *obj, ConfigOpts opts)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    char *errmsg;

    g_assert(core->config);

    configuration_set_opts(core->config, opts);

    if (!configuration_save(core->config, &errmsg)) {
        nntpgrab_core_emit_fatal_error(errmsg);
        g_free(errmsg);
    }
}

void
nntpgrab_schedular_initialize(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular == NULL);

    core->schedular = schedular_new(core->config);
}

gboolean
nntpgrab_schedular_start(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);

    return schedular_start(core->schedular);
}

static gpointer
schedular_stop_func(gpointer data)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(data);

    g_assert(core->schedular);

    schedular_stop(core->schedular);

    return NULL;
}

gboolean
nntpgrab_schedular_stop(NNTPGrabCore *obj, gboolean wait)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);

    if (wait) {
        return schedular_stop(core->schedular);
    } else {
        // Is the schedular queue already stopped?
        if (schedular_get_state(core->schedular) != SCHEDULAR_STATE_RUNNING) {
            return FALSE;
        }

        g_thread_create(schedular_stop_func, obj, FALSE, NULL);

        return TRUE;
    }
}

SchedularState
nntpgrab_schedular_get_state(NNTPGrabCore *obj)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);

    return (schedular_get_state(core->schedular));
}

gboolean
nntpgrab_schedular_add_task_to_queue(NNTPGrabCore *obj, const char *collection_name, const char *subject, const char *poster, time_t stamp, guint64 file_size, GList *groups, GList *parts, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);
    NNTPFile *file;
    GList *list;
    char *tmp;
    char *tmp2;
    ConfigOpts opts;

    g_assert(core->schedular);

    file = g_malloc0(sizeof(NNTPFile));
    strncpy((char *) file->subject, subject, sizeof(file->subject));
    strncpy((char *) file->poster, poster, sizeof(file->poster));
    file->stamp = stamp;
    file->file_size = file_size;

    opts = configuration_get_opts(core->config);
    tmp = tempnam(opts.temp_directory, NULL);
    tmp2 = g_path_get_basename(tmp);
    strncpy((char *) file->tmp_filename, tmp2, sizeof(file->tmp_filename));
    g_free(tmp2);
    free(tmp);                  // Don't use g_free here as tempnam is a libc function!

    list = groups;
    while (list) {
        const char *group = list->data;
        file->groups = g_list_append(file->groups, g_strdup(group));
        list = g_list_next(list);
    }

    list = parts;
    while (list) {
        NNTPGrabPart *part = list->data;
        NNTPPart *new_part;

        new_part = g_malloc0(sizeof(NNTPPart));

        strncpy((char *) new_part->message_id, part->message_id, sizeof(new_part->message_id));
        new_part->size = part->size;
        new_part->partnum = part->partnum;

        file->parts = g_list_append(file->parts, new_part);

        list = g_list_next(list);
    }

    // FIXME: This needs to be calculated out of the full subjects
    file->numparts = g_list_length(parts);

    return schedular_add_task_to_queue(core->schedular, collection_name, file, errmsg);
}

gboolean
nntpgrab_schedular_del_task_from_queue(NNTPGrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);
    g_assert(collection_name);

    return schedular_del_task_from_queue(core->schedular, collection_name, subject, errmsg);
}

gboolean
nntpgrab_schedular_restart_task(NNTPGrabCore *obj, const char *collection_name, const char *subject, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);

    return schedular_restart_task(core->schedular, collection_name, subject, errmsg);
}

gboolean
nntpgrab_schedular_save_queue(NNTPGrabCore *obj, char **errmsg)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);

    return schedular_save_queue(core->schedular, errmsg);
}

void
nntpgrab_schedular_foreach_task(NNTPGrabCore *obj, ForeachCollectionFunc collection_func, ForeachFileFunc file_func, ForeachGroupFunc group_func, gpointer data)
{
    NNTPGrabCore *core = NNTPGRAB_CORE(obj);

    g_assert(core->schedular);

    schedular_foreach_task(core->schedular, collection_func, file_func, group_func, data);
}
